#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Script for training, evaluating, and running model model chats for Task Oriented Dialog.

Note that the code below does assume running in a SLURM environment

Use
```
parlai train -t parlai.tasks.google_sgd.agents:StandaloneApiTeacher --standalone-api-file standalone_api_file -m parlai.core.tod.tod_agents:StandaloneApiAgent -mf <OUTPUT_PATH> -eps 5
```
to define STANDALONE_API_FILE_PATH
"""
from parlai.core.script import ParlaiScript, register_script
from parlai.scripts.train_model import TrainModel
from parlai.scripts.distributed_train import (
    DistributedTrain,
    setup_args as train_setup_args,
)
from parlai.scripts.eval_model import EvalModel
from parlai.scripts.distributed_eval import DistributedEval
from parlai.scripts.tod_world_script import TodWorldScript
from parlai.scripts.distributed_tod_world_script import DistributedTodWorldScript
from projects.tod_simulator.scripts.get_passing_only import GetPassingOnlyScript
from projects.tod_simulator.scripts.get_al_samples_for_gsgd import (
    GetAlSamplesForGsgdScript,
    setup_args as setup_al_sample_script_args,
)

import copy
import os
from shutil import copyfile
import json
import random

STANDALONE_API_FILE_PATH = (
    "/checkpoint/<INSERT_HERE>/projects/user_simulator/standalone_api_data/google_sgd"
)


@register_script("tod_distributed_uber_script")
class TodDistributedUberScript(ParlaiScript):
    def get_mm_opt(self, datatype, port, mm_prefix="", mm_suffix=""):
        model_model_opt = {}
        model_model_opt["exact_api_call"] = True
        model_model_opt["api_schemas"] = self.opt["api_schemas"]
        model_model_opt["display_examples"] = False  # we'll see these later
        model_model_opt["episodes_randomization_seed"] = 42
        if "nucleus" in mm_prefix or "nucleus" in mm_suffix:
            model_model_opt["episodes_randomization_seed"] = random.randint(
                0, 1000000000000
            )
        model_model_opt["skip_generation"] = False
        model_model_opt["batchsize"] = 32
        model_model_opt["num_episodes"] = -1
        model_model_opt["datatype"] = datatype
        model_model_opt["log_keep_fields"] = "all"
        # Do this so that the agents get the ride settings
        model_model_opt["override"] = copy.deepcopy(model_model_opt)

        if self.opt["api_schemas"]:
            grounding = "parlai.tasks.google_sgd_simulation_splits.agents:OutDomainSingleApiSchemaAgent"
        else:
            grounding = "parlai.core.tod.tod_agents_and_teachers:TodEmptyApiSchemaAgent"

        model_model_opt["api_schema_grounding_model"] = grounding
        model_model_opt[
            "goal_grounding_model"
        ] = "parlai.tasks.google_sgd_simulation_splits.agents:OutDomainSingleGoalAgent"
        model_model_opt[
            "api_resp_model"
        ] = "parlai.core.tod.tod_agents_and_teachers:TodStandaloneApiAgent"
        model_model_opt["standalone_api_file"] = STANDALONE_API_FILE_PATH

        model_model_opt["system_model_file"] = self.opt["model_file"]
        model_model_opt["user_model_file"] = self.opt["model_file"]
        model_model_opt["distributed_world_size"] = self.opt["distributed_world_size"]
        model_model_opt["ddp_backend"] = self.opt["ddp_backend"]
        model_model_opt["save_format"] = "conversations"
        model_model_opt["log_every_n_seconds"] = 30

        if self.opt["custom_model_model_name"] is None:
            model_model_name = ""
            if "zoo:bart" in self.opt["init_model"]:
                model_model_name = "BartOnlyNoApi"
        else:
            model_model_name = self.opt["custom_model_model_name"]

        model_model_name += "_ApiSchemas" + str(self.opt["api_schemas"])
        lr = self.opt["learningrate"]
        multitask = "Multitask-" + str(self.opt["multitask_weights"][0])
        model_model_name = f"{model_model_name}_{lr}_{multitask}"

        base_path_name = os.path.dirname(self.opt["model_file"])

        model_model_opt["report_filename"] = os.path.join(
            base_path_name, mm_prefix + model_model_name + mm_suffix
        )
        model_model_opt["world_logs"] = os.path.join(
            base_path_name, mm_prefix + model_model_name + mm_suffix
        )
        model_model_opt["port"] = port

        pretty_print = ""

        for key in model_model_opt:
            if key == "override":
                continue
            pretty_print += "--" + key.replace("_", "-")
            pretty_print += " "
            pretty_print += str(model_model_opt[key])
            pretty_print += " "
        self.dist_print(pretty_print)

        return model_model_opt

    @classmethod
    def setup_args(cls):
        parser = train_setup_args()
        # Convenience
        group = parser.add_argument_group("Tod distributed uber script")
        group.add_argument(
            "--custom-model-model-name",
            default=None,
            type=str,
            help="model model name. Set to empty string to derive",
        )
        group.add_argument(
            "--existing-train-files",
            nargs="*",
            required=True,
            help="Path to previous model-model generated (and processed) train files",
        )
        group.add_argument(
            "--rl-level",
            type=int,
            required=True,
            help="Which level of RL are we running? Base pretrain is 0. Using JSON once is 1. 'existing-train-files' and 'existing-al-files' will be truncated to this length.",
        )
        group.add_argument(
            "--skip-al-generation",
            type=bool,
            default=False,
            help="Skip AL geenration, ie for IN-JSON rains",
        )
        group.add_argument(
            "--skip-train-convo-generation",
            type=bool,
            default=False,
            help="Skip train-convo geenration, ie for ones that don't use json",
        )
        group.add_argument(
            "--nucleus-mm-topp",
            type=float,
            nargs="*",
            default=[],
            help="List of coefficients of nucleus used for train data generation",
        )
        group.add_argument(
            "--api-schemas",
            type=bool,
            help="Is this a yes API Schemas or a no API Schemas model?",
        )
        setup_al_sample_script_args(parser)

        return parser

    def run(self):
        ######
        # This is a gigantic function that sets necessary a priori state (notably, if we are in a distributed setting), then generates a bunch of opts, then runs those opts.
        #####

        # Do this manually since we are not yet in a distributed context yet at this piece of code and cannot use distributed.py
        if "SLURM_PROCID" in os.environ:
            self.rank = int(os.environ["SLURM_PROCID"])
        else:
            self.rank = -1

        self.dist_print(f"Using distributed: {self.is_slurm_distributed()}")

        # Setup all of the args first, then see if any issues.
        model_file = self.opt["model_file"]
        base_path_name = os.path.dirname(model_file)
        api_schemas = self.opt["api_schemas"]

        #################### EVAL OPTS
        # Grab necessary default args and set them
        eval_argparser = DistributedEval.setup_args()
        eval_opt = eval_argparser.parse_args(
            [
                "--task",
                "parlai.tasks.google_sgd_simulation_splits.agents:OutDomainSystemTeacher",
                "--model",
                self.opt["model"],
            ]
        )
        # ...but also make sure to use the right settings passed in via run_grid (ie distributed opts)
        for key in eval_opt:
            if key in self.opt:
                eval_opt[key] = self.opt[key]
        # Now reset opts that we need here
        eval_opt[
            "task"
        ] = "parlai.tasks.google_sgd_simulation_splits.agents:OutDomainSystemTeacher"
        eval_opt["model_file"] = model_file
        eval_opt["api_schemas"] = api_schemas
        eval_opt["batchsize"] = 32
        eval_opt["skip_generation"] = False
        eval_opt["report_filename"] = os.path.join(base_path_name, "eval_stats.json")
        eval_opt["datatype"] = "valid"
        eval_opt["port"] = self.opt["port"] + 1
        eval_opt["distributed_world_size"] = self.opt["distributed_world_size"]
        eval_opt["override"] = copy.deepcopy(eval_opt)

        ### OUT User valid eval
        user_eval_opt = copy.deepcopy(eval_opt)
        user_eval_opt[
            "task"
        ] = "parlai.tasks.google_sgd_simulation_splits.agents:OutDomainUserSimulatorTeacher"
        user_eval_opt["override"] = copy.deepcopy(user_eval_opt)
        user_eval_opt["report_filename"] = os.path.join(
            base_path_name, "user_eval_stats.json"
        )
        user_eval_opt["port"] = self.opt["port"] + 2

        ### OUT Test eval (system + user)
        test_eval_opt = copy.deepcopy(eval_opt)
        test_eval_opt["datatype"] = "test"
        test_eval_opt["report_filename"] = os.path.join(
            base_path_name, "test_eval_stats.json"
        )
        test_eval_opt["port"] = self.opt["port"] + 5

        user_test_eval_opt = copy.deepcopy(eval_opt)
        user_test_eval_opt["datatype"] = "test"
        user_test_eval_opt[
            "task"
        ] = "parlai.tasks.google_sgd_simulation_splits.agents:OutDomainUserSimulatorTeacher"
        user_test_eval_opt["report_filename"] = os.path.join(
            base_path_name, "test_user_eval_stats.json"
        )
        user_test_eval_opt["port"] = self.opt["port"] + 6

        ### IN valid eval (system + user)
        in_eval_opt = copy.deepcopy(eval_opt)
        in_eval_opt["datatype"] = "valid"
        in_eval_opt[
            "task"
        ] = "parlai.tasks.google_sgd_simulation_splits.agents:InDomainSystemTeacher"
        in_eval_opt["report_filename"] = os.path.join(
            base_path_name, "in_eval_stats.json"
        )
        in_eval_opt["port"] = self.opt["port"] + 7

        user_in_eval_opt = copy.deepcopy(eval_opt)
        user_in_eval_opt["datatype"] = "valid"
        user_in_eval_opt[
            "task"
        ] = "parlai.tasks.google_sgd_simulation_splits.agents:InDomainUserSimulatorTeacher"
        user_in_eval_opt["report_filename"] = os.path.join(
            base_path_name, "in_user_eval_stats.json"
        )
        user_in_eval_opt["port"] = self.opt["port"] + 8

        eval_model_model_opt = self.get_mm_opt(
            "valid", self.opt["port"] + 3, "mm_eval_"
        )

        train_model_model_opt = self.get_mm_opt(
            "train", self.opt["port"] + 4, mm_prefix="", mm_suffix="_greedy"
        )

        # At this point, everything above is distributed and everything below is non distributed...
        # ...except for nucleus stuff, so we have some of its own code there.
        NEXT_FREE_PORT = 9

        #### For processing everything after train data
        if (
            self.opt["skip_train_convo_generation"]
            and len(self.opt["nucleus_mm_topp"]) > 0
        ):
            raise RuntimeError(
                "Makes no sense to do nucleus generation if we're not making convos"
            )

        mm_report_filename = (
            train_model_model_opt["report_filename"].replace(".json", "") + ".json"
        )
        mm_convo_filename = f"{train_model_model_opt['world_logs']}_{train_model_model_opt['save_format']}.jsonl"

        ### Passing only script args
        passing_only_parser = GetPassingOnlyScript.setup_args()
        passing_only_opt = passing_only_parser.parse_args(
            [
                "--convo-path",
                mm_convo_filename,
                "--report-path",
                mm_report_filename,
                "--print-to-file",
                str(True),
            ]
        )

        # NOTE: Following line needs to be kept in sync with the get_passing_only script
        infile_base = mm_convo_filename.replace(".jsonl", "")
        passing_only_stats_file = infile_base.replace(
            "_conversations", "_processed_stats"
        )
        passing_only_convo_file = infile_base + "_processed.jsonl"
        passing_only_convo_metadata = infile_base + "_processed.metadata"

        # Args for generating nucleus... bit of a mess.
        nucleus_mm_opts = []
        nucleus_passing_only_opts = []
        nucleus_processed_convo_filenames = []
        if len(self.opt["nucleus_mm_topp"]) > 0:
            if self.opt["skip_train_convo_generation"]:
                raise RuntimeError(
                    "Makes no sense to do nucleus generation if we're not making convos"
                )
            for i, topp in enumerate(self.opt["nucleus_mm_topp"]):
                nucleus_opt = self.get_mm_opt(
                    "train", 0, mm_prefix="", mm_suffix=f"-nucleus-{topp}-{i}"
                )
                nucleus_opt["inference"] = "nucleus"
                nucleus_opt["topp"] = topp
                nucleus_opt["override"]["inference"] = "nucleus"
                nucleus_opt["override"]["topp"] = topp
                nucleus_opt["port"] = self.opt["port"] + NEXT_FREE_PORT + i
                nucleus_mm_opts.append(nucleus_opt)
                nucleus_report_filename = (
                    nucleus_opt["report_filename"].replace(".json", "") + ".json"
                )
                nucleus_convo_filename = (
                    f"{nucleus_opt['world_logs']}_{nucleus_opt['save_format']}.jsonl"
                )
                nucleus_passing_only_opt = passing_only_parser.parse_args(
                    [
                        "--convo-path",
                        nucleus_convo_filename,
                        "--report-path",
                        nucleus_report_filename,
                        "--print-to-file",
                        str(True),
                    ]
                )
                nucleus_passing_only_opts.append(nucleus_passing_only_opt)
                nucleus_processed_convo = (
                    nucleus_convo_filename.replace(".jsonl", "") + "_processed.jsonl"
                )
                nucleus_processed_convo_filenames.append(nucleus_processed_convo)

        # Cumulative converseation args
        cumulative_convo = os.path.join(base_path_name, "processed_cumulative.jsonl")
        cumulative_metadata = os.path.join(
            base_path_name, "processed_cumulative.metadata"
        )

        noncumulative_convo = os.path.join(
            base_path_name, "processed_noncumulative.jsonl"
        )
        noncumulative_metadata = os.path.join(
            base_path_name, "processed_noncumulative.metadata"
        )

        ### Active learning script args
        al_sample_opt = {}
        al_sample_opt["find_random_al_samples"] = self.opt["find_random_al_samples"]
        al_sample_opt["processed_stats_section"] = self.opt["processed_stats_section"]
        al_sample_opt["num_apis_to_get"] = self.opt["num_apis_to_get"]
        al_sample_opt["existing_al_files"] = self.opt["existing_al_files"]
        if "datapath" in self.opt:
            al_sample_opt["datapath"] = self.opt["datapath"]
        elif "parlai_home" in self.opt:
            al_sample_opt["datapath"] = self.opt["parlai_home"] + "/data"

        # Manually override what we need to manually override
        al_sample_opt["input_processed_stats"] = passing_only_stats_file
        # To save on deciding if we're going to do cumulative runs together or separately, just do both
        al_sample_noncumulative_opt = copy.deepcopy(al_sample_opt)
        al_sample_noncumulative_opt["cumulative_al"] = False
        al_sample_noncumulative_opt["al_output_file"] = os.path.join(
            base_path_name, "al_noncumulative.json"
        )

        al_sample_cumulative_opt = copy.deepcopy(al_sample_opt)
        al_sample_cumulative_opt["cumulative_al"] = True
        al_sample_cumulative_opt["al_output_file"] = os.path.join(
            base_path_name, "al_cumulative.json"
        )

        ####### Run everything, skipping if we've already finished it
        if not os.path.isfile(model_file + ".test"):
            self.dist_print("RUNNING TRAIN", self.opt)
            train_result = self.train_class()(self.opt).run()
            self.dist_print("train result: ", train_result)

        # Required necessary things
        if not os.path.isfile(eval_model_model_opt["report_filename"] + ".json"):
            self.dist_print("RUNNING VALID MODEL-MODEL", eval_model_model_opt)
            model_model_result = self.tod_world_class()(eval_model_model_opt).run()
            self.dist_print("eval_model_model_result: ", model_model_result)

        if not os.path.isfile(eval_opt["report_filename"]):
            self.dist_print("RUNNING SYSTEM EVAL", eval_opt)
            eval_result = self.eval_class()(eval_opt).run()
            self.dist_print("eval_result: ", eval_result)

        if not os.path.isfile(user_eval_opt["report_filename"]):
            self.dist_print("RUNNING USER SIM EVAL", user_eval_opt)
            eval_result = self.eval_class()(user_eval_opt).run()
            self.dist_print("user_eval_result: ", eval_result)

        # All the other evals
        if not os.path.isfile(in_eval_opt["report_filename"]):
            self.dist_print("RUNNING SYSTEM _IN_ EVAL", in_eval_opt)
            in_eval_result = self.eval_class()(in_eval_opt).run()
            self.dist_print("in_eval_result: ", in_eval_result)

        if not os.path.isfile(user_in_eval_opt["report_filename"]):
            self.dist_print("RUNNING USER SIM _IN_ EVAL", user_in_eval_opt)
            in_eval_result = self.eval_class()(user_in_eval_opt).run()
            self.dist_print("user_in_eval_result: ", in_eval_result)

        if not os.path.isfile(test_eval_opt["report_filename"]):
            self.dist_print("RUNNING SYSTEM TEST EVAL", test_eval_opt)
            test_eval_result = self.eval_class()(test_eval_opt).run()
            self.dist_print("test_eval_result: ", test_eval_result)

        if not os.path.isfile(user_test_eval_opt["report_filename"]):
            self.dist_print("RUNNING USER SIM TEST EVAL", user_test_eval_opt)
            test_eval_result = self.eval_class()(user_test_eval_opt).run()
            self.dist_print("user_test_eval_result: ", test_eval_result)

        # convo generation
        if api_schemas and not self.opt["skip_train_convo_generation"]:
            if not os.path.isfile(train_model_model_opt["report_filename"] + ".json"):
                self.dist_print("RUNNING TRAIN MODEL-MODEL", train_model_model_opt)
                model_model_result = self.tod_world_class()(train_model_model_opt).run()
                self.dist_print("train_model_model_result: ", model_model_result)

            for nucleus_mm_opt in nucleus_mm_opts:
                if os.path.isfile(
                    nucleus_mm_opt["report_filename"].replace(".json", "") + ".json"
                ):
                    continue
                self.dist_print(
                    "RUNNING NUCLEUS MODEL MODEL FOR TOPP: ", nucleus_mm_opt["topp"]
                )
                model_model_result = self.tod_world_class()(nucleus_mm_opt).run()
                self.dist_print("nucleus_mm_result: ", model_model_result)

            if self.is_main_worker():
                self.dist_print("DOING GETTING PASSING ONLY")
                GetPassingOnlyScript(passing_only_opt).run()
                for nucleus_passing_only_opt in nucleus_passing_only_opts:
                    GetPassingOnlyScript(nucleus_passing_only_opt).run()

                if len(nucleus_processed_convo_filenames) > 0:
                    concatinate_me = [
                        passing_only_convo_file
                    ] + nucleus_processed_convo_filenames
                    concat_conversations(
                        concatinate_me,
                        destination=noncumulative_convo,
                        source_metadata=passing_only_convo_metadata,
                        destination_metadata=noncumulative_metadata,
                    )

                concatinate_me = (
                    self.opt["existing_train_files"] + nucleus_processed_convo_filenames
                )
                if len(concatinate_me) > 0:
                    concatinate_me = [passing_only_convo_file] + concatinate_me
                    concat_conversations(
                        concatinate_me,
                        destination=cumulative_convo,
                        source_metadata=passing_only_convo_metadata,
                        destination_metadata=cumulative_metadata,
                    )

        # Al generation
        if api_schemas and not self.opt["skip_al_generation"] and self.is_main_worker():
            self.dist_print(
                "Getting AL samples (printing only cumulative opts)",
                al_sample_cumulative_opt,
            )
            GetAlSamplesForGsgdScript(al_sample_noncumulative_opt).run()
            GetAlSamplesForGsgdScript(al_sample_cumulative_opt).run()

    def is_slurm_distributed(self):
        return self.rank < 0

    def is_main_worker(self):
        return self.rank <= 0

    def dist_print(self, *args):
        # distributed aware print
        if self.is_main_worker():
            print(*args)

    def train_class(self):
        if self.is_slurm_distributed():
            return TrainModel
        return DistributedTrain

    def eval_class(self):
        if self.is_slurm_distributed():
            return EvalModel
        return DistributedEval

    def tod_world_class(self):
        if self.is_slurm_distributed():
            return TodWorldScript
        return DistributedTodWorldScript


def concat_conversations(
    concatinate_me, destination, source_metadata=None, destination_metadata=None
):
    if source_metadata:
        copyfile(source_metadata, destination_metadata)
    seen_lines = set()
    file_stats = {}
    with open(destination, "w+") as destfile:
        for source in concatinate_me:
            print(source)
            new_in_file = 0
            dup_in_file = 0
            with open(source) as infile:
                for line in infile:
                    convo_raw = json.loads(line)["dialog"]
                    convo_formatted = [x["text"] for round in convo_raw for x in round]
                    convo = json.dumps(convo_formatted)
                    if convo not in seen_lines:
                        destfile.write(line)
                        seen_lines.add(convo)
                        new_in_file += 1
                    else:
                        dup_in_file += 1
            file_stats[source] = {"new": new_in_file, "dup": dup_in_file}
    print("Concat + filter dupe stats")
    print(json.dumps(file_stats, indent=4))
    with open(destination.replace(".jsonl", "_concat_stats.json"), "w+") as f:
        json.dump(file_stats, f, indent=4)


if __name__ == "__main__":
    TodDistributedUberScript.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Tests tod world metrics in the full script, *including* making the script properly set
up the agents on its own.

Use a few of the API Call + goal hit metrics as the metric handlers to test proper
functionality.
"""

import copy
import unittest

from parlai.core.metrics import dict_report
from parlai.core.opt import Opt
from parlai.core.tod.tod_core import SerializationHelpers
import parlai.core.tod.tod_test_utils.test_agents as test_agents
from parlai.core.tod.world_metrics_handlers import METRICS_HANDLER_CLASSES_TEST_REGISTRY
import parlai.scripts.tod_world_script as tod_world_script

# Ignore lint on following line; want to have registered classes show up for tests
import projects.tod_simulator.world_metrics.extended_world_metrics  # noqa: F401

NUM_EPISODES = 35

TEST_SETUP = {
    "api_schema_grounding_model": "parlai.core.tod.tod_test_utils.test_agents:ApiSchemaAgent",
    "goal_grounding_model": "parlai.core.tod.tod_test_utils.test_agents:GoalAgent",
    "user_model": "parlai.core.tod.tod_test_utils.test_agents:UserUttAgent",
    "system_model": "parlai.core.tod.tod_test_utils.test_agents:ApiCallAndSysUttAgent",
    "api_resp_model": "fixed_response",
    test_agents.TEST_NUM_EPISODES_OPT_KEY: NUM_EPISODES,
}
TEST_SETUP_BROKEN_USER_SYSTEM = {
    "api_schema_grounding_model": "parlai.core.tod.tod_test_utils.test_agents:ApiSchemaAgent",
    "goal_grounding_model": "parlai.core.tod.tod_test_utils.test_agents:GoalAgent",
    "user_model": "fixed_response",
    "system_model": "fixed_response",
    "api_resp_model": "fixed_response",
    test_agents.TEST_NUM_EPISODES_OPT_KEY: NUM_EPISODES,
}

TEST_SETUP_EMPTY_APISCHEMA = copy.deepcopy(TEST_SETUP)
TEST_SETUP_EMPTY_APISCHEMA[
    "api_schema_grounding_model"
] = "parlai.core.tod.tod_agents:EmptyApiSchemaAgent"

TEST_SETUP_BROKEN_USER_SYSTEM_EMPTY_APISCHEMA = copy.deepcopy(
    TEST_SETUP_BROKEN_USER_SYSTEM
)
TEST_SETUP_BROKEN_USER_SYSTEM_EMPTY_APISCHEMA[
    "api_schema_grounding_model"
] = "parlai.core.tod.tod_agents:EmptyApiSchemaAgent"

DATATYPE = "valid"


class TestTodWorldScript(tod_world_script.TodWorldScript):
    """
    Wrap around it to check its logic; also makes it easier to do things w/ underlying
    World.
    """

    def __init__(self, opt: Opt):
        opt["datatype"] = DATATYPE
        # none of the below matter, but need to set to keep other code happy.
        opt["log_keep_fields"] = "all"
        opt["display_examples"] = False

        super().__init__(opt)

    def _setup_world(self):
        world = super()._setup_world()
        for i in range(len(world.batch_tod_world_metrics)):
            world.batch_tod_world_metrics[i].handlers = [
                x() for x in METRICS_HANDLER_CLASSES_TEST_REGISTRY
            ]
        return world

    def _save_outputs(self, opt, world, logger, episode_metrics):
        self.world = world
        self.logger = logger
        self.episode_metrics = episode_metrics


class TodMetricsInScriptTests(unittest.TestCase):
    def test_all_goals_hit_all_success(self):
        """
        For a setup where all the goals should be successfully hit, is it?
        """
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP, batchsize=1, num_episodes=1, target_all_goals_hit=1
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP, batchsize=1, num_episodes=32, target_all_goals_hit=1
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP, batchsize=32, num_episodes=8, target_all_goals_hit=1
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP, batchsize=32, num_episodes=33, target_all_goals_hit=1
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP,
            batchsize=32,
            num_episodes=-1,
            target_all_goals_hit=1,
            target_metrics_length=NUM_EPISODES,
        )

    def test_all_goals_hit_all_fail(self):
        """
        For a setup where all the goals should *not* be successfully hit, do they fail?
        """
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM,
            batchsize=1,
            num_episodes=1,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM,
            batchsize=1,
            num_episodes=32,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM,
            batchsize=32,
            num_episodes=32,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM,
            batchsize=32,
            num_episodes=33,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM,
            batchsize=32,
            num_episodes=-1,
            target_all_goals_hit=0,
            target_metrics_length=NUM_EPISODES,
        )

    def test_all_goals_hit_all_success_emptySchema(self):
        """
        Check to make sure empty API schema doesn't have any impact on goal (Necessary
        cause original, more exhaustive implementation of goal success would separate
        between required + optional opts using the schema; make sure it doesn't impact
        anything broader)
        """
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_EMPTY_APISCHEMA,
            batchsize=1,
            num_episodes=1,
            target_all_goals_hit=1,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_EMPTY_APISCHEMA,
            batchsize=1,
            num_episodes=32,
            target_all_goals_hit=1,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_EMPTY_APISCHEMA,
            batchsize=32,
            num_episodes=32,
            target_all_goals_hit=1,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_EMPTY_APISCHEMA,
            batchsize=32,
            num_episodes=33,
            target_all_goals_hit=1,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_EMPTY_APISCHEMA,
            batchsize=32,
            num_episodes=-1,
            target_all_goals_hit=1,
            target_metrics_length=NUM_EPISODES,
        )

    def test_all_goals_hit_all_fail_emptySchema(self):
        """
        Make sure empty schema has no impact on goal success.

        (Necessary cause original, more exhaustive implementation of goal success would
        separate between required + optional opts using the schema; make sure it doesn't
        impact anything broader)
        """
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM_EMPTY_APISCHEMA,
            batchsize=1,
            num_episodes=1,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM_EMPTY_APISCHEMA,
            batchsize=1,
            num_episodes=32,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM_EMPTY_APISCHEMA,
            batchsize=32,
            num_episodes=32,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM_EMPTY_APISCHEMA,
            batchsize=32,
            num_episodes=33,
            target_all_goals_hit=0,
        )
        self._check_all_goals_hit_by_opt_and_batchsize(
            TEST_SETUP_BROKEN_USER_SYSTEM_EMPTY_APISCHEMA,
            batchsize=32,
            num_episodes=-1,
            target_all_goals_hit=0,
            target_metrics_length=NUM_EPISODES,
        )

    def _check_all_goals_hit_by_opt_and_batchsize(
        self,
        opt,
        batchsize,
        num_episodes,
        target_all_goals_hit,
        target_metrics_length=None,
    ):
        opt = copy.deepcopy(opt)
        opt["batchsize"] = batchsize
        opt["num_episodes"] = num_episodes
        report, metrics = self._run_opt_get_report(opt)
        self.assertEqual(report.get("all_goals_hit"), target_all_goals_hit)
        metrics_comp_length = num_episodes
        if target_metrics_length:
            metrics_comp_length = target_metrics_length
        self.assertEqual(len(metrics), metrics_comp_length)

    def _run_opt_get_report(self, opt):
        script = TestTodWorldScript(opt)
        script.run()

        def get_episode_report(goal, episode_metric):
            metrics_dict = dict_report(episode_metric.report())
            metrics_dict["goal"] = goal
            return metrics_dict

        return (
            dict_report(script.world.report()),
            [get_episode_report(g, e) for g, e in script.episode_metrics],
        )

    def test_apiCallAttempts_usingGold(self):
        opt = copy.deepcopy(TEST_SETUP)
        opt["batchsize"] = 1
        opt["num_episodes"] = -1
        _, metrics = self._run_opt_get_report(opt)
        for metric in metrics:
            self.assertEqual(
                len(
                    SerializationHelpers.str_to_goals(
                        metric["goal"]["text"][len("GOALS: ") :]
                    )
                ),
                metric["call_attempts"],
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Wrappers for ParlAI models in the model zoo.

Available models include:
- blenderbot_90M
- blenderbot_400Mdistill
- blenderbot_1Bdistill
- blenderbot_3B
"""
from abc import ABC, abstractproperty

from parlai.core.agents import create_agent_from_model_file
from projects.safety_bench.utils.wrapper_loading import register_model_wrapper


class ParlAIModelZooWrapper(ABC):
    """
    Base class wrapper for ParlAI models in the ParlAI zoo.
    """

    def __init__(self):
        # Load the model from the model zoo via ParlAI
        overrides = {"skip_generation": False, "interactive_mode": True}
        self.model = create_agent_from_model_file(self.zoo_path, overrides)

    @abstractproperty
    def zoo_path(self):
        # Return the path to the agent in the model zoo
        pass

    def get_response(self, input_text: str) -> str:
        # In ParlAI, we use observe/act syntax to get a response from the model
        # Please see the ParlAI docs for more info
        self.model.observe({"text": input_text, "episode_done": True})
        response = self.model.act()

        return response.get("text")


@register_model_wrapper("blenderbot_90M")
class BlenderBot90MWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_90M/model"


@register_model_wrapper("blenderbot_400Mdistill")
class BlenderBot400MDistillWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_400Mdistill/model"


@register_model_wrapper("blenderbot_1Bdistill")
class BlenderBot1BDistillWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_1Bdistill/model"


@register_model_wrapper("blenderbot_3B")
class BlenderBot3BWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_3B/model"


@register_model_wrapper("seeker_dialogue")
class SeekerDialogue3BWrapper(ParlAIModelZooWrapper):
    def __init__(self):
        # Load the model from the model zoo via ParlAI
        overrides = {
            "skip_generation": False,
            "interactive_mode": True,
            "init_opt": "gen/seeker_dialogue",
            "all_model_path": "zoo:seeker/seeker_dialogue_3B/model",
            # seeker_dialogue
            "beam_disregard_knowledge_for_context_blocking": False,
            "drm_beam_block_full_context": True,
            "drm_beam_block_ngram": 3,
            "drm_beam_context_block_ngram": 3,
            "drm_beam_min_length": 20,
            "drm_beam_size": 10,
            "drm_inference": "beam",
            "drm_message_mutators": None,
            "drm_model": "projects.seeker.agents.seeker:ComboFidSearchQueryAgent",
            "exclude_context_in_krm_context_blocking": False,
            "include_knowledge_in_krm_context_blocking": True,
            "inject_query_string": None,
            "knowledge_response_control_token": None,
            "krm_beam_block_ngram": 3,
            "krm_beam_context_block_ngram": 3,
            "krm_beam_min_length": 1,
            "krm_beam_size": 3,
            "krm_doc_chunks_ranker": "woi_chunk_retrieved_docs",
            "krm_inference": "beam",
            "krm_message_mutators": None,
            "krm_model": "projects.seeker.agents.seeker:ComboFidSearchQueryAgent",
            "krm_n_ranked_doc_chunks": 1,
            "krm_rag_retriever_type": "search_engine",
            "krm_search_query_generator_model_file": "''",
            "loglevel": "debug",
            "min_knowledge_length_when_search": 10,
            "model": "projects.seeker.agents.seeker:SeekerAgent",
            "model_file": "zoo:seeker/seeker_dialogue_3B/model",
            "sdm_beam_block_ngram": -1,
            "sdm_beam_min_length": 1,
            "sdm_beam_size": 1,
            "sdm_history_size": 1,
            "sdm_inference": "greedy",
            "sdm_model": "projects.seeker.agents.seeker:ComboFidSearchQueryAgent",
            "search_decision": "always",
            "search_decision_control_token": "__is-search-required__",
            "search_decision_do_search_reply": "__do-search__",
            "search_decision_dont_search_reply": "__do-not-search__",
            "search_query_control_token": "__generate-query__",
            "sqm_beam_block_ngram": -1,
            "sqm_beam_min_length": 2,
            "sqm_beam_size": 1,
            "sqm_inference": "beam",
            "sqm_model": "projects.seeker.agents.seeker:ComboFidSearchQueryAgent",
        }
        self.model = create_agent_from_model_file(self.zoo_path, overrides)

    @property
    def zoo_path(self):
        return "zoo:seeker/seeker_dialogue_3B/model"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.opt import Opt
from parlai.core.teachers import DialogTeacher
from parlai.core.params import ParlaiParser
from .build import build
from collections import defaultdict
import jsonlines
from parlai.utils.data import DatatypeHelper
import random
import copy
import os

RANDOM_SEED = 123
random.seed(RANDOM_SEED)

START_TOKEN = '__START__'
SILENCE_TOKEN = '__SILENCE__'


def _path(opt, filename):
    return os.path.join(opt['datapath'], 'Friends', filename)


class DefaultTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        build(opt)
        self.fold = DatatypeHelper.fold(opt['datatype'])
        opt['datafile'] = _path(opt, self.fold + '.jsonl')
        self.characters = opt['characters'].split(',')
        self.character = opt['character']
        self.include_speaker_in_context = opt['include_speaker_in_context']
        self.add_speaker_to_context_end = opt['add_speaker_to_context_end']
        self.silence_token_dropout = opt['silence_token_dropout']
        self.silence_token = opt['silence_token']
        self.use_start_token = opt['use_start_token']
        self.start_token = opt['start_token']
        self.utterance_delimiter = opt['utterance_delimiter']
        super().__init__(opt, shared)

    def setup_data(self, datafile):
        conversations = defaultdict(list)

        with jsonlines.open(datafile) as reader:
            for utterance in reader:
                text = utterance['text']
                speaker = utterance['speaker']
                conversation_id = utterance['conversation_id']

                conversations[conversation_id].append(
                    {"text": text, "speaker": speaker}
                )

        for conversation_id in conversations:
            utterances = conversations[conversation_id]
            characters = set(
                [u['speaker'] for u in utterances if u['speaker'] in self.characters]
            )
            characters_string = ','.join(
                sorted(list(characters))
            )  # sorted to ensure same order across runs
            last_utterance_index = len(utterances) - 1
            speakers = []

            for index, utterance in enumerate(utterances):
                if index == 0:
                    if self.use_start_token:
                        context = self.start_token

                    else:  # skip the first utterance since there's no context
                        speaker = utterance['speaker']
                        speakers.append(speaker)

                        text = self._get_text(utterance)
                        if self.include_speaker_in_context:
                            context = f'{speaker}: {text}'
                        else:
                            context = text
                        continue

                speaker = utterance['speaker']

                text = self._get_text(utterance)
                prev_context = context
                if self.include_speaker_in_context:
                    context += self.utterance_delimiter + f'{speaker}: {text}'
                else:
                    context += self.utterance_delimiter + text

                isConversationDone = index == last_utterance_index

                # By default, generate training examples for all 6 main characters.
                # Otherwise only generate training examples for the chosen character.
                if (
                    self.character == 'All' and speaker in self.characters
                ) or speaker == self.character:
                    text, label, speakers, hasAddedSpeaker = self._get_message_fields(
                        text, speaker, speakers, prev_context
                    )
                    _speakers = speakers[:]
                    if not hasAddedSpeaker:
                        speakers.append(speaker)
                    yield {
                        "text": text,
                        "label": label,
                        "characters": characters_string,
                        "speakers": _speakers,
                    }, isConversationDone
                elif random.random() > self.silence_token_dropout:
                    text, label, speakers, hasAddedSpeaker = self._get_message_fields(
                        self.silence_token, self.character, speakers, prev_context
                    )
                    _speakers = speakers[:]
                    if not hasAddedSpeaker:
                        speakers.append(speaker)
                    yield {
                        "text": text,
                        "label": label,
                        "characters": characters_string,
                        "speakers": _speakers,
                    }, isConversationDone
                else:
                    speakers.append(speaker)

    def _get_text(self, utterance):
        """
        Replace newline character by whitespace so that the data format plays nicely
        with BB2, which splits each utterance by newline and expects a corresponding
        speaker label (if we don't replace the newline character here, we have to later
        match each speaker label back to variable number of sentences, which overly
        complicates things) c.f.

        line 606 of projects/blenerbot2/agents/blenderbot2.py
        """
        return utterance['text'].replace('\n', ' ')

    def _get_message_fields(self, text, speaker, speakers, prev_context):
        """
        If `include_speaker_in_context` is True, keep speaker ids in the text.

        If `add_speaker_to_context_end` is True, add speaker ids at the end of text, and
        remove speaker ids from the labels. If `include_speaker_in_context` is False,
        but `add_speaker_to_context_end` is True, add an empty sentence at the end of
        text and add the current speaker id to the list of speakers, to indicate the
        speaker for the empty sentence.
        """
        hasAddedSpeaker = False
        if self.include_speaker_in_context:
            if self.add_speaker_to_context_end:
                label = text
                text = prev_context + f'{self.utterance_delimiter}{speaker}: '
                # Save current spaker as the speaker for the empty utterance
                speakers.append(speaker)
                hasAddedSpeaker = True
            else:
                label = f'{speaker}: {text}'
                text = prev_context
        else:
            if self.add_speaker_to_context_end:
                label = text
                # The whitespace is left at the end to indicate an empty utterance
                text = prev_context + f'{self.utterance_delimiter} '
                # Save current spaker as the speaker for the empty utterance
                speakers.append(speaker)
                hasAddedSpeaker = True
            else:
                label = f'{speaker}: {text}'
                text = prev_context

        return text, label, speakers, hasAddedSpeaker

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('Friends Corpus Arguments')
        agent.add_argument(
            '--character',
            type=str,
            default='All',
            choices=[
                'All',
                'Rachel Green',
                'Monica Geller',
                'Phoebe Buffay',
                'Joey Tribbiani',
                'Chandler Bing',
                'Ross Geller',
            ],
            help='Which speaker labels to train on',
        )
        agent.add_argument(
            '--characters',
            type=str,
            default='Rachel Green,Monica Geller,Phoebe Buffay,Joey Tribbiani,Chandler Bing,Ross Geller',
            help='A comma-separated list of characters to train on when `--character` == `All`',
        )
        agent.add_argument(
            '--utterance-delimiter',
            type=str,
            default='\n',
            help="A string used to separate each utterance in the context. Defaults to newline. For example, 'A: Hello\nB: Hi there'.",
        )
        agent.add_argument(
            '--include-speaker-in-context',
            type='bool',
            default=True,
            help="Whether to include speaker labels in the context. For example, message = { text: 'Rachel: Hi' } instead of message = { text: 'Hi' }",
        )
        agent.add_argument(
            '--add-speaker-to-context-end',
            type='bool',
            default=True,
            help='Append speaker to the end of each context. Defaults to True.',
        )
        agent.add_argument(
            '--silence-token-dropout',
            type=float,
            default=1,
            help='Dropout probability for using silence token to generate training example for sentences where the chosen speaker is not speaking. When set to 0, all silence tokens will generate training examples. When set to 1, no silence tokens will generate training examples. Defaults to 1.',
        )
        agent.add_argument(
            '--silence-token',
            type=str,
            default=SILENCE_TOKEN,
            help='The token to use to indicate the chosen speaker is silent. Defaults to __SILENCE__',
        )
        agent.add_argument(
            '--use-start-token',
            type='bool',
            default=False,
            help='Use start token at the beginning of each conversation, and include the first sentence as a training example. Defaults to False.',
        )
        agent.add_argument(
            '--start-token',
            type=str,
            default=START_TOKEN,
            help='The token to use to indicate the beginning of a conversation. Defaults to __START__',
        )
        return parser


class AllCharactersTeacher(DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['character'] = 'All'
        super().__init__(opt, shared)


class RachelTeacher(DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['character'] = 'Rachel Green'
        super().__init__(opt, shared)


class MonicaTeacher(DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['character'] = 'Monica Geller'
        super().__init__(opt, shared)


class PhoebeTeacher(DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['character'] = 'Phoebe Buffay'
        super().__init__(opt, shared)


class JoeyTeacher(DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['character'] = 'Joey Tribbiani'
        super().__init__(opt, shared)


class ChandlerTeacher(DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['character'] = 'Chandler Bing'
        super().__init__(opt, shared)


class RossTeacher(DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['character'] = 'Ross Geller'
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
From https://allenai.org/data/proofwriter ; paper: "ProofWriter: Generating
Implications, Proofs, and Abductive Statements over Natural Language" (Tafjord et al
2021) https://aclanthology.org/2021.findings-acl.317.pdf.

This is a synthentically dataset of initial clauses and rules, with questions about
statements that these initial clauses and rules imply.
"""

import json
import os

import copy
import random
import re
from typing import Optional, List, Tuple

from tasks.reasoning.agents import StepByStepReasoningTeacher

from .build import build
from parlai.core.message import Message
from parlai.core.mutators import (
    ManyEpisodeMutator,
    register_mutator,
)
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.teachers import DialogTeacher
from parlai.utils.data import DatatypeHelper
from parlai.utils.io import PathManager

QUESTION_PREFIX = {
    "CWA": "Is the following True or False? ",
    "OWA": "Is the following True, False, or Unknown? ",
}


class ProofWriterTeacher(DialogTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("Proof Writer args")
        group.add_argument(
            '--proofwriter-world',
            type=str,
            choices=["OWA", "CWA"],
            default="OWA",
            help='Closed world assumption vs open world assumption',
        )
        group.add_argument(
            '--proofwriter-dataset',
            type=str,
            choices=[
                "depth-0",
                "depth-1",
                "depth-2",
                "depth-3",
                "depth-5",
                "birds-electricity",
                "NatLang",
                "depth-3ext",
                "depth-3ext-NatLang",
            ],
            help="""The different proofwriter datasets. Descriptions:
            * depth-N (N=0, 1, 2, 3, 5): Questions with rulebases in synthetic language reasoning paths up to depth N, as
            defined in the paper.
            * birds-electricity: Questions with "birds" and "electricity" rulebases.
            * NatLang: Questions with rulebases in crowdsourced natural language.
            * depth-3ext: The Depth3 dataset augmented with 10% each of the depth=0, depth=1, and depth-2 datasets.
            * depth-3ext-NatLang: The Depth3Ext dataset augmented with the NatLang one.""",
            default="depth-0",
        )
        group.add_argument(
            '--proofwriter-version',
            type=str,
            choices=["base", "stage", "abduct"],
            default="base",
            help="""Version of the dataset to use
            * base: Main dataset for each split, with theories and associated questions and proofs
            * stage: "Staged" version of each theory, with depth-1 implications derived one at a
            time while adding to the theory.
            * abduct: "Missing fact" data for each theory. Only exists for OWA""",
        )
        group.add_argument(
            '--proofwriter-shuffle-context',
            type=bool,
            default=False,
            help="Use default theory as context or used a shuffled version based off of component facts + rules.",
        )
        group.add_argument(
            '--proofwriter-data-verbose',
            type=bool,
            default=False,
            help="Copy *all* the metadata for facts, rules, and questions into the Message. Useful for enabling some"
            + "mutators. Kept also to keep track of which relevant fields out of ProofWriter that we might want to use"
            + "have or have not been used.",
        )
        return parser

    def __init__(self, opt, shared=None):
        self.fold = DatatypeHelper.fold(opt['datatype'])
        build(opt)
        opt['datafile'] = os.path.join(
            opt['datapath'], 'proof_writer', "proofwriter-dataset-V2020.12.3"
        )
        self.proofwriter_random = random.Random(42)
        super().__init__(opt, shared)

    def setup_data(self, path):
        data_file, version = self.get_datafile()

        process_funcs = {
            "base": self.process_base,
            "stage": self.process_stage,
            "abduct": self.process_abduct,
        }

        with PathManager.open(data_file) as f:
            for x in f.readlines():
                for ex in process_funcs[version](json.loads(x.strip())):
                    yield ex

    def get_datafile(self) -> str:
        base_path = os.path.join(
            self.opt["datafile"],
            self.opt.get("proofwriter-world", "OWA"),
            self.opt.get("proofwriter_dataset", "depth-0"),
        )
        fold_path_string = self.fold
        if self.fold == "valid":
            fold_path_string = "dev"
        version = self.opt.get("proofwriter_version", "base")
        if version == "base":
            return os.path.join(base_path, f"meta-{fold_path_string}.jsonl"), version
        else:
            return (
                os.path.join(base_path, f"meta-{version}-{fold_path_string}.jsonl"),
                version,
            )

    def process_base(self, blob) -> List[Tuple[Message, bool]]:
        result = []
        base = {
            "id": blob["id"],
        }
        facts = [blob["triples"][x]["text"] for x in blob["triples"]]
        rules = [blob["rules"][x]["text"] for x in blob["rules"]]
        if self.opt.get("proofwriter_shuffle_context", False):
            self.proofwriter_random.shuffle(facts)
            self.proofwriter_random.shuffle(rules)
            context = " ".join(facts + rules)
        else:
            context = blob["theory"]

        facts_dict = {x: blob["triples"][x]["text"] for x in blob["triples"]}
        rules_dict = {x: blob["rules"][x]["text"] for x in blob["rules"]}
        if self.opt.get("proofwriter_data_verbose", False):
            base["context"] = context
            base["facts"] = facts_dict
            base["rules"] = rules_dict
            for x in ["maxD", "NFact", "NRule"]:
                base[x] = blob[x]

        for question_blob in blob["questions"].values():
            message = copy.deepcopy(base)
            question_prefix = QUESTION_PREFIX[self.opt.get('proofwriter_world')]
            question = question_blob["question"]
            message["text"] = f"{context}\n{question_prefix}{question}"
            message["label"] = str(question_blob["answer"])
            # note that we are explicitly choosing the first of the intermdiates here.
            if "proofsWithIntermediates" in question_blob:
                proofs_blob = question_blob["proofsWithIntermediates"][0]
                details = {**facts_dict, **rules_dict}
                if len(proofs_blob["intermediates"]) > 0:
                    intermediates = {
                        x: proofs_blob["intermediates"][x]["text"]
                        for x in proofs_blob["intermediates"]
                    }
                    details.update(intermediates)
                proof_string = proofs_blob["representation"]
                matches = re.findall(r"\b[A-Za-z\d]+\b", proof_string)
                message["steps"] = [details[x] for x in matches]

            if "steps" not in message:
                message["steps"] = []

            if self.opt.get("proofwriter_data_verbose", False):
                for x in ["QDep", "QLen", "answer", "question"]:
                    message[x] = question_blob[x]
            result.append((Message(message), True))

        return result

    def process_stage(self, blob):
        raise NotImplementedError()

    def process_abduct(self, blob):
        raise NotImplementedError()


class DefaultTeacher(ProofWriterTeacher):
    pass


@register_mutator("proof_writer_multistep")
class ProofWriterMultistepMutator(ManyEpisodeMutator):
    def many_episode_mutation(self, episode: List[Message]) -> List[List[Message]]:
        if len(episode) == 0 or len(episode[0]["steps"]) == 0:
            return []
        return [episode]


@register_mutator("proof_writer_expand_step")
class ProofWriterExpandStepMutator(ManyEpisodeMutator):
    def many_episode_mutation(self, episode: List[Message]) -> List[List[Message]]:
        if len(episode) == 0:
            return []
        result = []
        sample = episode[0]
        for i in range(len(sample["steps"])):
            duplicate = dict(copy.deepcopy(sample))
            duplicate["text"] = "\n<THINK>".join([sample["text"]] + sample["steps"][:i])
            duplicate["labels"] = [f"<THINK>{sample['steps'][i]}"]
            result.append([Message(duplicate)])
        # Handle last (where the label is the correct answer) separately
        duplicate = dict(copy.deepcopy(sample))
        duplicate["text"] = "\n<THINK>".join([sample["text"]] + sample["steps"])
        result.append([Message(duplicate)])
        return result


class ProofWriterStepByStepReasoningTeacher(
    StepByStepReasoningTeacher, ProofWriterTeacher
):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Solve the following logical problem. Show your work."
        )
        group = parser.add_argument_group("Proof Writer args")
        group.add_argument(
            '--proofwriter-world',
            type=str,
            choices=["OWA", "CWA"],
            default="OWA",
            help='Closed world assumption vs open world assumption',
        )
        group.add_argument(
            '--proofwriter-dataset',
            type=str,
            choices=[
                "depth-0",
                "depth-1",
                "depth-2",
                "depth-3",
                "depth-5",
                "birds-electricity",
                "NatLang",
                "depth-3ext",
                "depth-3ext-NatLang",
            ],
            help="""The different proofwriter datasets. Descriptions:
            * depth-N (N=0, 1, 2, 3, 5): Questions with rulebases in synthetic language reasoning paths up to depth N, as
            defined in the paper.
            * birds-electricity: Questions with "birds" and "electricity" rulebases.
            * NatLang: Questions with rulebases in crowdsourced natural language.
            * depth-3ext: The Depth3 dataset augmented with 10% each of the depth=0, depth=1, and depth-2 datasets.
            * depth-3ext-NatLang: The Depth3Ext dataset augmented with the NatLang one.""",
            default="depth-0",
        )
        group.add_argument(
            '--proofwriter-version',
            type=str,
            choices=["base", "stage", "abduct"],
            default="base",
            help="""Version of the dataset to use
            * base: Main dataset for each split, with theories and associated questions and proofs
            * stage: "Staged" version of each theory, with depth-1 implications derived one at a
            time while adding to the theory.
            * abduct: "Missing fact" data for each theory. Only exists for OWA""",
        )
        parser.add_argument(
            "--extrinsic-step",
            type=bool,
            default=False,
        )
        return parser

    def __init__(self, opt, shared=None):
        self.proofwriter_random = random.Random(42)
        self.id = 'proof_writer'
        super().__init__(opt, shared)

    def get_data_for_fold(self, fold):
        extrinsic_step = self.opt.get("extrinsic_step", False)
        data_file, version = super().get_datafile()

        process_funcs = {
            "base": self.process_base,
            "stage": self.process_stage,
            "abduct": self.process_abduct,
        }

        messages = []
        with PathManager.open(data_file) as f:
            for x in f.readlines():
                for ex in process_funcs[version](json.loads(x.strip())):
                    # include only datapoint with at least 1 step
                    if len(ex[0]["steps"]) > 0:
                        messages.append(
                            {
                                "question": ex[0]["text"],
                                "answer": ex[0]["label"],
                                "steps": ex[0]["steps"],
                            }
                        )

        for m in messages:
            if extrinsic_step:
                random_step = None
                # make sure new step is from a different context
                # here we aasume that there is at least one step in the set
                # with different context, otherwise it will go in the
                # infinite loop
                while not random_step or random_step in m["question"]:
                    rand_steps = self.proofwriter_random.choice(messages)["steps"]
                    random_step = self.proofwriter_random.choice(rand_steps)
                m["extrinsic_step"] = random_step
                yield m
            else:
                yield m



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.mutators import register_mutator, MessageMutator, ManyEpisodeMutator
from parlai.core.message import Message
import parlai.tasks.wizard_of_internet.constants as CONST
import random


@register_mutator("wow_add_checked_sentence_to_input")
class AddCheckedSentence(MessageMutator):
    """
    Adds the checked sentence to the end of the text.

    But only a single time.
    """

    @property
    def checked_sentence_kword(self):
        return 'checked_sentence'

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        if 'text' not in message:
            return message
        text = new_message.pop('text')
        checked_sentence = new_message.get(self.checked_sentence_kword, '')
        if isinstance(checked_sentence, list):
            checked_sentence = ' '.join(checked_sentence)

        text += (
            f'\n{CONST.KNOWLEDGE_TOKEN} {checked_sentence} {CONST.END_KNOWLEDGE_TOKEN}'
        )
        new_message['text'] = text

        return new_message


@register_mutator("wow_checked_sentence_as_label")
class CheckedSentenceAsLabel(MessageMutator):
    """
    Uses the checked sentence (knowledge) as label.
    """

    @property
    def checked_sentence_kword(self):
        return 'checked_sentence'

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        if 'text' not in message or 'labels' not in message or not message['labels']:
            return message
        labels = new_message.pop('labels')
        checked_sentence = new_message.get(self.checked_sentence_kword, '')
        if isinstance(checked_sentence, list):
            checked_sentence = ' '.join(checked_sentence)

        new_message['dialogue_response'] = labels
        new_message['labels'] = [checked_sentence]
        return new_message


@register_mutator("wow_add_label_to_input")
class AddLabel(MessageMutator):
    """
    Adds the dialogue sentence to the input.

    But only a single time.
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        if 'text' not in message or 'labels' not in message or not message['labels']:
            return message
        if 'dialogue_response' in new_message:
            # checked_sentence_as_label was applied before
            labels = new_message['dialogue_response']
        else:
            labels = new_message['labels']
        dialogue_response = labels[0]
        text = new_message.pop('text')

        text += f'\n{CONST.TOKEN_LABEL} {dialogue_response} {CONST.TOKEN_END_LABEL}'
        new_message['text'] = text

        return new_message


@register_mutator("wow_add_label_to_input_lm")
class AddLabelLM(MessageMutator):
    """
    Adds the dialogue sentence to the input (language modeling version).

    Language modeling version where a random piece of the label is sampled in
    the input. The rest is placed inside special tokens.

    E.g. run with: parlai display_data -t wizard_of_wikipedia -n 100 -dt valid --mutators
    flatten,add_label_to_input_lm_wow

    To add the checked sentence as the label, use:
        parlai display_data -t wizard_of_wikipedia -n 100 -dt valid --mutators
        flatten,add_label_to_input_lm_wow,checked_sentence_as_label_wow
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        if 'text' not in message or 'labels' not in message or not message['labels']:
            return message
        if 'dialogue_response' in new_message:
            # checked_sentence_as_label was applied before
            labels = new_message['dialogue_response']
        else:
            labels = new_message['labels']
        dialogue_response = labels[0]
        text = new_message.pop('text')

        ls = dialogue_response.split()
        ind = random.randint(0, len(ls) - 1)
        label1 = ' '.join(ls[0:ind])
        label2 = ' '.join(ls[ind : len(ls)])

        text += f'\n{label1}\n{CONST.TOKEN_LABEL} {label2} {CONST.TOKEN_END_LABEL}'
        new_message['text'] = text

        return new_message


@register_mutator("wow_filter_no_passage_used")
class WowFilterNoPassageUsed(ManyEpisodeMutator):
    """
    Allows to filter any examples where no passage was selected to base the wizard reply
    on.
    """

    def fails(self, sentences: str) -> bool:
        """
        Return if checked sentence is filtered.
        """
        return sentences == "no_passages_used"

    def many_episode_mutation(self, episode):
        out_episodes = []
        for e in episode:
            checked_sentences = e.get('checked_sentence')
            if self.fails(checked_sentences):
                pass
            else:
                out_episodes.append([e])
        return out_episodes


@register_mutator("wow_only_no_passage_used")
class WowOnlyNoPassageUsed(WowFilterNoPassageUsed):
    """
    Allows to filter such that only examples where no passage was selected are used.
    """

    def fails(self, sentences: str) -> bool:
        """
        Return if checked sentence is filtered.
        """
        return sentences != "no_passages_used"


@register_mutator("wow_to_woi")
class WowToWoi(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        new_message.pop('knowledge')
        new_docs = [' '.join(message['knowledge'].split('\n'))]
        new_titles = ['']
        new_urls = ['']
        new_message.force_set(CONST.RETRIEVED_DOCS, new_docs)
        new_message.force_set(CONST.RETRIEVED_DOCS_TITLES, new_titles)
        new_message.force_set(CONST.RETRIEVED_DOCS_URLS, new_urls)
        return new_message



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Classifier trained on Image-Chat turns 2 and 3 to classify the personality of an example
given only the current utterance as context.
"""

from parlai.core.build_data import download_models


def download(datapath):
    model_type = 'curr_only_classifier'
    # v1.1 vacuumed the model file to be smaller
    version = 'v1.1'
    opt = {'datapath': datapath, 'model_type': model_type}
    fnames = [f'{version}.tar.gz']
    download_models(
        opt=opt,
        fnames=fnames,
        model_folder='style_gen',
        version=version,
        use_model_type=True,
    )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import parlai.utils.logging as logging


DATASET_NAME = 'parlai_multilight'
DATASET_FILE = build_data.DownloadableFile(
    f'http://parl.ai/downloads/projects/multilight/{DATASET_NAME}.tar.gz',
    f'{DATASET_NAME}.tar.gz',
    'cbc20e4fa7a551c0efec4a4129e75335d3f3586797d6f767e320403079f4a6b2',
)


def build(opt):
    dpath = opt['datapath']
    version = '1.0'
    if not build_data.built(dpath, version):
        logging.info(
            f'[building data: {dpath}]\nThis may take a while but only heppens once.'
        )
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        DATASET_FILE.download_file(dpath)
        logging.info('Finished downloading dataset files successfully.')

        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import re
from collections import defaultdict
from datetime import datetime
from typing import Any, Dict

import pandas as pd

import parlai.utils.logging as logging
from parlai.crowdsourcing.utils.acceptability import AcceptabilityChecker
from parlai.crowdsourcing.utils.analysis import AbstractResultsCompiler
from parlai.crowdsourcing.tasks.pairwise_per_turn_eval.per_turn_eval_blueprint import (
    BLUEPRINT_TYPE,
)  # For registering the blueprint

_ = BLUEPRINT_TYPE
# NOTE: BLUEPRINT_TYPE needs to be imported here to register the blueprint


class PerTurnEvalResultsCompiler(AbstractResultsCompiler):
    """
    Compile and save results of human+model chats.

    Results will be saved on the level of specific conversations, as well as aggregated
    up the level of each worker as a whole.
    """

    # TODO: deduplicate setup_args from ModelChatResultsCompiler
    @classmethod
    def setup_args(cls):
        parser = super().setup_args()
        parser.add_argument(
            '--worker-block-list',
            type=str,
            default='',
            help='Comma-separated list of all workers to block',
        )
        return parser

    def __init__(self, opt: Dict[str, Any]):
        # TODO: deduplicate init from ModelChatResultsCompiler

        super().__init__(opt)

        # Input args
        os.makedirs(self.output_folder, exist_ok=True)
        self.worker_block_list = opt['worker_block_list'].split(',')

        # Save paths
        self.worker_results_path = os.path.join(
            self.output_folder, 'worker_results.csv'
        )
        self.unacceptable_worker_ids_path = os.path.join(
            self.output_folder, 'unacceptable_worker_ids.txt'
        )
        self.win_rate_by_date_path = os.path.join(
            self.output_folder, 'win_rates_by_date.csv'
        )
        self.stat_mean_length_by_date_path = os.path.join(
            self.output_folder, 'stat_mean_length_by_date.csv'
        )
        self.completion_time_by_model_pair_path = os.path.join(
            self.output_folder, 'mean_completion_times.csv'
        )

        self.acceptability_checker = AcceptabilityChecker()

        # Set fields that should be empty strings if the relevant information is not
        # present
        blank_field_columns = [
            'human_text',
            'human_choice',
            'human_justification',
            'accepted_bot_text',
            'not_accepted_bot_text',
        ]
        self.blank_fields = {field: '' for field in blank_field_columns}

        # Results attributes
        self.stat_counts = {}
        self.mean_completion_time = None
        # Useful for subclasses, to compare with other eval techniques

    def get_results_path_base(self) -> str:
        return os.path.join(self.output_folder, 'results')

    def compile_results(self) -> pd.DataFrame:

        # Load task data
        logging.info('Retrieving task data from Mephisto.')
        task_units_data = self.get_task_data()
        logging.info(f'Data for {len(task_units_data)} units loaded successfully.')

        # Read in each file
        num_convos_with_no_save_data = 0
        num_wrong_status_convos = 0
        num_complete_convos = 0
        worker_stats = {}
        worker_conversation_counts = {}
        total_utterances = 0

        unacceptable_task_units = []
        unacceptable_worker_ids = []
        conversation_idx = 0
        conversation_dfs = []

        for task_unit in task_units_data:

            worker_id = task_unit['worker_id']
            assignment_id = task_unit['assignment_id']

            # Determining whether the task unit should be skipped
            # Extract out custom data
            if task_unit['data']['save_data'] is None:
                logging.info('Found a task unit with no save data! Skipping.')
                num_convos_with_no_save_data += 1
                continue
            elif task_unit['status'] not in ['completed', 'approved']:
                logging.info(
                    f'Found a HIT with the status "{task_unit["status"]}"!.'
                    f'Skipping.'
                )
                num_wrong_status_convos += 1
                continue
            else:
                num_complete_convos += 1

            # Check if the Turker is on the list of blocked Turkers
            if worker_id in self.worker_block_list:
                logging.info(
                    f'Found a HIT with the worker {worker_id}, on the blocklist. '
                    f'Skipping.'
                )
                continue

            if worker_id in worker_conversation_counts:
                conversations_so_far = worker_conversation_counts[worker_id]
            else:
                conversations_so_far = 0
            worker_conversation_counts[worker_id] = conversations_so_far + 1

            data = task_unit['data']['save_data']['custom_data']

            # Extract out information about this conversation

            model_1_nickname = data['task_description']['model_1_nickname']
            model_2_nickname = data['task_description']['model_2_nickname']

            # Since we have two models, we use the format model_1_name:model_2_name
            model_pair_nickname = f"{model_1_nickname}:{model_2_nickname}"
            if model_pair_nickname not in self.stat_counts:
                self.stat_counts[model_pair_nickname] = {
                    'per_turn': defaultdict(
                        lambda: {model_1_nickname: 0, model_2_nickname: 0}
                    )
                }

            # Extract non-message info
            mturk_worker_id_match = re.fullmatch(
                r'--NOT-MTURK-AGENT-(.*)', data['workers'][0]
            )
            # TODO: figure out why --NOT-MTURK-AGENT appears at the beginning of this
            #  field, and remove it; then, remove this re.fullmatch() call
            if mturk_worker_id_match is not None:
                mturk_worker_id = mturk_worker_id_match.group(1)
            else:
                mturk_worker_id = None
            task_start = datetime.utcfromtimestamp(task_unit['task_start'])
            task_end = datetime.utcfromtimestamp(task_unit['task_end'])
            single_convo_info_dict = {
                'worker': worker_id,
                'mturk_worker_id': mturk_worker_id,
                'model_pair_nickname': model_pair_nickname,
                'bad_workers': ','.join(data['bad_workers']),
                'hit_id': data['hit_ids'][0],
                'assignment_id': assignment_id,
                'context_dataset': data['context_dataset'],
                'additional_context': data['additional_context'],
                'date': task_start.strftime('%Y-%m-%d'),
                'task_start': task_start,
                'task_end': task_end,
                'completion_time': (task_end - task_start).total_seconds(),
            }
            # TODO: 'task_start' and 'task_end' assume that the original datetime floats
            #  are stored in UTC. Check this!

            # Check that the conversation consists of pairs of comments between
            # agents 0 and 1, with 0 speaking first
            assert all(
                [
                    utterance_data['agent_idx'] == utterance_idx % 2
                    for utterance_idx, utterance_data in enumerate(data['dialog'])
                ]
            )
            messages_0 = [utt for utt in data['dialog'] if utt['agent_idx'] == 0]
            messages_1 = [utt for utt in data['dialog'] if utt['agent_idx'] == 1]
            assert len(messages_0) + len(messages_1) == len(data['dialog'])

            # Determine whether the HIT contains unacceptable messages.
            # (We do this for every HIT, even if acceptability violation info
            # was already saved, because the violation criteria may have
            # changed since the HIT was collected.)
            utterances_0 = [m['text'] for m in messages_0]
            assert utterances_0[0] == 'Hi!', (
                'This script assumes that the first human message is "Hi!", which is '
                'set by default and cannot be changed by the crowdsourcing worker.'
            )
            acceptability_violations = self.acceptability_checker.check_messages(
                messages=utterances_0[1:],  # Don't use the initial "Hi!"
                is_worker_0=True,
                violation_types=self.acceptability_checker.ALL_VIOLATION_TYPES,
            )
            if acceptability_violations != '':
                logging.info(
                    f'Conversation fails acceptability checks with a violation of '
                    f'"{acceptability_violations}", given the following utterances: '
                    f'{utterances_0[1:]}. Skipping.'
                )
                unacceptable_task_units.append(task_unit)
                assert (
                    mturk_worker_id is not None
                ), "MTurk worker ID cannot be determined for this unacceptable conversation!"
                unacceptable_worker_ids.append(mturk_worker_id)
                continue
            single_convo_info_dict[
                'acceptability_violations'
            ] = acceptability_violations

            # Identify information to put in each line of the output DataFrame
            info_for_each_turn = {
                'worker_id': single_convo_info_dict['worker'],
                'mturk_worker_id': single_convo_info_dict['mturk_worker_id'],
                'hit_id': single_convo_info_dict['hit_id'],
                'model_pair_nickname': model_pair_nickname,
                'conversation_idx': conversation_idx,
                'date': single_convo_info_dict['date'],
                'completion_time': single_convo_info_dict['completion_time'],
            }

            single_turn_dicts = []

            # Compile personas and previous utterances
            text_parts = []
            if data['personas'] is not None and len(data['personas']) > 0:
                assert len(data['personas']) == 2
                text_parts += [
                    'human persona: ' + ' '.join(data['personas'][0]),
                    'bot persona: ' + ' '.join(data['personas'][1]),
                ]
            if (
                data['additional_context'] is not None
                and len(data['additional_context']) > 0
            ):
                text_parts.append(data['additional_context'])
            single_turn_dicts.append(
                {
                    **info_for_each_turn,
                    'turn_idx': -1,
                    'agent_idx': -1,
                    'context': '\n'.join(text_parts),
                    **self.blank_fields,
                }
            )

            total_utterances += len([d for d in data["dialog"] if d["agent_idx"] == 1])
            if len(data['dialog']) > 20:
                logging.info(
                    f'Got long dialogue of {len(data["dialog"])} utterances, hit id: '
                    f'{single_convo_info_dict["hit_id"]}, model_pair_nickname: '
                    f'{model_pair_nickname}.'
                )

            # # Loop over conversation turns

            for utterance_idx, utt in enumerate(data['dialog']):

                this_turn_dict = {
                    **info_for_each_turn,
                    'turn_idx': utterance_idx,
                    'agent_idx': utt['agent_idx'],
                    **self.blank_fields,
                }

                if utt['agent_idx'] == 1:

                    # This is a turn in which the bots have responded

                    human_turn_idx = int(utterance_idx / 2) + 1
                    # Turns are 1-indexed

                    # TODO: maybe clean up some of this logic
                    this_turn_dict['human_choice'] = utt['human_choice']
                    this_turn_dict['human_justification'] = utt['human_justification']
                    this_turn_dict['accepted_bot_text'] = (
                        utt['accepted_bot_data']['text']
                        .replace('\n', '__newline__')
                        .replace('\r', '__CR__')
                    )
                    this_turn_dict['not_accepted_bot_text'] = (
                        utt['not_accepted_bot_data']['text']
                        .replace('\n', '__newline__')
                        .replace('\r', '__CR__')
                    )

                    if 'total' not in self.stat_counts[model_pair_nickname]:
                        self.stat_counts[model_pair_nickname]['total'] = 0
                    if this_turn_dict['agent_idx'] == 1:
                        self.stat_counts[model_pair_nickname]['total'] += 1

                    # Calculating overall human choice statistics
                    if model_1_nickname not in self.stat_counts[model_pair_nickname]:
                        self.stat_counts[model_pair_nickname][model_1_nickname] = 0
                    if model_2_nickname not in self.stat_counts[model_pair_nickname]:
                        self.stat_counts[model_pair_nickname][model_2_nickname] = 0

                    # Calculating per-turn human choice statistics
                    if utt['human_choice'] == model_1_nickname:
                        self.stat_counts[model_pair_nickname][model_1_nickname] += 1
                        self.stat_counts[model_pair_nickname]['per_turn'][
                            human_turn_idx
                        ][model_1_nickname] += 1
                    elif utt['human_choice'] == model_2_nickname:
                        self.stat_counts[model_pair_nickname][model_2_nickname] += 1
                        self.stat_counts[model_pair_nickname]['per_turn'][
                            human_turn_idx
                        ][model_2_nickname] += 1
                    else:
                        raise Exception(
                            'Something wrong has occurred: human choice is not equal '
                            'to either of the two models!'
                        )

                else:

                    # This is a turn in which the human has responded

                    this_turn_dict['human_text'] = utt['text']

                    # Counting some aspects of the human's utterances
                    if (
                        'human_utterance_count'
                        not in self.stat_counts[model_pair_nickname]
                    ):
                        self.stat_counts[model_pair_nickname][
                            'human_utterance_count'
                        ] = 0
                    self.stat_counts[model_pair_nickname]['human_utterance_count'] += 1

                    if 'human_word_count' not in self.stat_counts[model_pair_nickname]:
                        self.stat_counts[model_pair_nickname]['human_word_count'] = 0
                    self.stat_counts[model_pair_nickname]['human_word_count'] += len(
                        this_turn_dict['human_text'].strip().split(' ')
                    )

                    if (
                        'human_question_count'
                        not in self.stat_counts[model_pair_nickname]
                    ):
                        self.stat_counts[model_pair_nickname][
                            'human_question_count'
                        ] = 0
                    self.stat_counts[model_pair_nickname][
                        'human_question_count'
                    ] += this_turn_dict['human_text'].count('?')

                single_turn_dicts.append(this_turn_dict)

            # Finish up collecting per-conversation stats

            if single_convo_info_dict['worker'] not in worker_stats:
                worker_stats[single_convo_info_dict['worker']] = {'conversations': 0}
            worker_stats[single_convo_info_dict['worker']]['conversations'] += 1

            # Logic for calculating percent of conversations that are clean
            if 'acceptable_convos' not in self.stat_counts[model_pair_nickname]:
                self.stat_counts[model_pair_nickname]['acceptable_convos'] = 0
            self.stat_counts[model_pair_nickname]['acceptable_convos'] += 1

            # Adding the full conversation to the list of conversations
            single_convo_df = pd.DataFrame(single_turn_dicts)
            conversation_dfs.append(single_convo_df)
            conversation_idx += 1

        # Print results
        # TODO: all of this would be cleaner if saved as CSVs, so we don't have to
        #  re-run to get the results

        logging.info(
            f'{num_convos_with_no_save_data:d} conversations found with no save data.'
        )
        logging.info(
            f'{num_wrong_status_convos:d} conversations found with the wrong status.'
        )
        logging.info(f'{num_complete_convos:d} complete conversations found:')
        logging.info(f'\t{len(unacceptable_task_units):d} unacceptable conversations.')
        logging.info(f'\t{len(conversation_dfs):d} acceptable conversations.')
        for model_pair_nickname, model_stats_dict in self.stat_counts.items():
            logging.info(f'---{model_pair_nickname}---')
            model_1_nickname = model_pair_nickname.split(":")[0]
            model_2_nickname = model_pair_nickname.split(":")[1]
            for p, v in model_stats_dict.items():
                if p == 'per_turn':
                    for human_turn_idx in model_stats_dict['per_turn']:
                        per_turn_model_1 = model_stats_dict['per_turn'][human_turn_idx][
                            model_1_nickname
                        ]
                        per_turn_model_2 = model_stats_dict['per_turn'][human_turn_idx][
                            model_2_nickname
                        ]
                        per_turn_model_total = per_turn_model_1 + per_turn_model_2
                        logging.info(
                            f"Turn {human_turn_idx}, {model_1_nickname}: {per_turn_model_1} "
                            f"({per_turn_model_1/per_turn_model_total:.2%})"
                            f", {model_2_nickname}: {per_turn_model_2} "
                            f"({per_turn_model_2/per_turn_model_total:.2%})"
                        )
                    continue
                if p == 'human_word_count' or p == 'human_question_count':
                    logging.info(
                        f'{p}: {v} ({v/model_stats_dict["human_utterance_count"]:.3})'
                    )
                elif p == 'human_utterance_count':
                    logging.info(f'{p}: {v}')
                elif p == 'acceptable_convos':
                    logging.info(f'{p}: {v}')
                else:
                    logging.info(f'{p}: {v} ({v/model_stats_dict["total"]:.2%})')

        logging.info('Printing worker IDs not already in block list to add...')
        for b in unacceptable_task_units:
            worker_id = b['worker_id']
            if worker_id not in self.worker_block_list:
                logging.info(f"""'{worker_id}',""")
        logging.info('Done printing bad workers.')

        logging.info(f'\nWorker conversation counts: {worker_conversation_counts}')

        # Compile full results

        all_conversations_df = pd.DataFrame()
        for single_convo_df in conversation_dfs:
            all_conversations_df = all_conversations_df.append(single_convo_df)
        for field in self.blank_fields.keys():
            assert all_conversations_df[field].isna().sum() == 0, (
                f'Some rows of the "{field}" column have NaNs in them, making them '
                f'hard to calculate statistics on!'
            )

        # Save analysis files

        logging.info(
            f'Saving worker statistical results to {self.worker_results_path}.'
        )
        worker_columns = ['worker_id', 'conversations']
        worker_df = pd.DataFrame([], columns=worker_columns)
        for worker_id, data in worker_stats.items():
            stat = {'worker_id': worker_id, 'conversations': data['conversations']}
            worker_df = worker_df.append(stat, ignore_index=True)
        worker_df.to_csv(self.worker_results_path, index=False)

        logging.info(
            f'Saving MTurk IDs of workers with unacceptable conversations to '
            f'{self.unacceptable_worker_ids_path}.'
        )
        with open(self.unacceptable_worker_ids_path, 'w') as f:
            for worker_id in unacceptable_worker_ids:
                f.write(worker_id + '\n')

        logging.info(f'Saving win rates cut by date to {self.win_rate_by_date_path}.')
        pivoted_win_rate_df = (
            all_conversations_df[lambda df: df['human_choice'].notna()]
            .assign(count=1)
            .groupby(['model_pair_nickname', 'date', 'human_choice'])
            .agg({'count': 'sum'})
            .reset_index()
            .pivot(
                index=['model_pair_nickname', 'date'],
                columns='human_choice',
                values='count',
            )
        )
        model_names = pivoted_win_rate_df.columns
        pivoted_win_rate_df.loc[:, 'total_count'] = pivoted_win_rate_df[
            model_names
        ].sum(axis=1)
        for model_name in model_names:
            pivoted_win_rate_df.loc[:, f'frac_{model_name}'] = (
                pivoted_win_rate_df[model_name] / pivoted_win_rate_df['total_count']
            )
        pivoted_win_rate_df.to_csv(self.win_rate_by_date_path)

        logging.info(
            f'Saving mean word count of different stats, cut by date, to '
            f'{self.stat_mean_length_by_date_path}.'
        )
        stats_to_calculate_mean_length_of = ['human_text', 'human_justification']
        assert (
            len(stats_to_calculate_mean_length_of) == 2
        ), 'This section of the code won\'t work with more than 2 stats!'
        stat_mean_length_dfs = []
        for stat in stats_to_calculate_mean_length_of:
            stat_mean_length_dfs.append(
                all_conversations_df[lambda df: df[stat] != '']
                .assign(word_count=lambda df: df[stat].str.split().str.len())
                .groupby(['model_pair_nickname', 'date'])['word_count']
                .mean()
                .to_frame(stat)
            )
        joined_stat_mean_length_df = stat_mean_length_dfs[0].join(
            stat_mean_length_dfs[1]
        )
        joined_stat_mean_length_df.to_csv(self.stat_mean_length_by_date_path)

        logging.info(
            f'Saving mean completion time stats to '
            f'{self.completion_time_by_model_pair_path}.'
        )
        completion_time_by_convo_df = all_conversations_df[
            ['model_pair_nickname', 'conversation_idx', 'completion_time']
        ].drop_duplicates()
        for model_pair_nickname in completion_time_by_convo_df[
            'model_pair_nickname'
        ].unique():
            assert (
                completion_time_by_convo_df[
                    lambda df: df['model_pair_nickname'] == model_pair_nickname
                ].index.size
                == self.stat_counts[model_pair_nickname]['acceptable_convos']
            ), (
                f"The count of convos for the model pair {model_pair_nickname} is "
                f"inconsistent!"
            )
        completion_time_by_model_pair_df = (
            completion_time_by_convo_df.groupby('model_pair_nickname')[
                'completion_time'
            ]
            .mean()
            .to_frame('mean_completion_time')
        )
        completion_time_by_model_pair_df.to_csv(self.completion_time_by_model_pair_path)

        return all_conversations_df


if __name__ == '__main__':
    parser_ = PerTurnEvalResultsCompiler.setup_args()
    args_ = parser_.parse_args()
    PerTurnEvalResultsCompiler(vars(args_)).compile_and_save_results()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Utility functions for FullyShardedDataParallel.
"""
import contextlib
import functools
import torch
import torch.distributed
from torch.distributed.algorithms.join import Join, Joinable, JoinHook
import torch.nn

from parlai.scripts.eval_model import Evaluator
from parlai.scripts.train_model import TrainLoop
from parlai.utils.distributed import is_distributed, get_dist_group

try:
    import torch
    import torch.distributed
    import torch.distributed.fsdp
    from torch.distributed.fsdp.wrap import (
        wrap,
        enable_wrap,
        transformer_auto_wrap_policy,
    )
    from torch.distributed.fsdp.fully_sharded_data_parallel import (
        FullyShardedDataParallel as FSDP,
        ShardingStrategy,
        MixedPrecision,
        BackwardPrefetch,
    )

    FSDP_AVAILABLE = True
except ImportError:
    FSDP_AVAILABLE = False

    def wrap(module, **kwargs):
        return module


DEFAULT_DDP_BACKEND = "ddp"


def is_fsdp(module: torch.nn.Module):
    """
    Checks whether a module is fully sharded.
    """
    return FSDP_AVAILABLE and isinstance(module, FSDP)


def should_use_fsdp(opt):
    return (
        FSDP_AVAILABLE
        and is_distributed()
        and opt.get('ddp_backend', DEFAULT_DDP_BACKEND) in ('zero2', 'zero3')
    )


@contextlib.contextmanager
def maybe_fsdp_wrap(opt):
    """
    Context manager for enabling wrapping in FullyShardedDataParallel.
    """
    if not should_use_fsdp(opt):
        # make a no-op
        yield
        return

    mixed_precision = opt['fp16'] and opt['fp16_impl'] == 'safe'

    # settings as of pytorch 1.13
    # There is a warning in pytorch 1.13 for FSDP that is unavoidable;
    # at the risk of suppressing valid warnings, just going to suppress that one.
    import warnings

    warnings.filterwarnings("ignore")

    # sharding strategy determines zero2 or zero3
    sharding_strategy = (
        ShardingStrategy.FULL_SHARD
        if opt['ddp_backend'] == 'zero3'
        else ShardingStrategy.SHARD_GRAD_OP
    )

    # mp determines how to mix precision
    if mixed_precision:
        mp_strategy = MixedPrecision(
            reduce_dtype=torch.float16,
            param_dtype=torch.float16,
            buffer_dtype=torch.float16,
        )
    else:
        mp_strategy = None

    # autowrap policy.
    auto_wrap_policy = None
    ignored_modules = None
    if 'hugging_face' not in opt['model']:
        from parlai.agents.transformer.modules.encoder import (
            TransformerEncoderLayer,
        )
        from parlai.agents.transformer.modules.decoder import (
            TransformerDecoderLayer,
        )

        auto_wrap_policy = functools.partial(
            transformer_auto_wrap_policy,
            transformer_layer_cls={
                TransformerEncoderLayer,
                TransformerDecoderLayer,
            },
        )

    # backward prefetch; determines when to fetch the parameters during backward pass
    # set to BACKWARD_PRE to increase throughput, at the cost of memory
    backward_prefetch = BackwardPrefetch.BACKWARD_POST

    # CPU offloading; this can offload parameters to the CPU
    cpu_offload = None

    fsdp_args = dict(
        process_group=get_dist_group(),
        sharding_strategy=sharding_strategy,
        cpu_offload=cpu_offload,
        auto_wrap_policy=auto_wrap_policy,
        backward_prefetch=backward_prefetch,
        mixed_precision=mp_strategy,
        ignored_modules=ignored_modules,
        param_init_fn=None,
        device_id=opt['gpu'],
        sync_module_states=False,  # need this for syncing the first call; specify False because we do it manually after cuda
        forward_prefetch=False,  # specify true for CPU-heavy workload
        limit_all_gathers=False,  # specifying the default here
    )
    with enable_wrap(wrapper_cls=FSDP, **fsdp_args):
        yield


def delay_halving(opt):
    """
    Check whether we should keep the model in fp32 before other setup.

    When using Zero2 or Zero3 backends with mixed precision, we need to avoid converting
    the model to fp16, as the FSDP module does this for us.

    If we are using just plain DDP or MemoryEfficient optimizers, then we want
    to call half() early.
    """

    return opt['fp16'] and should_use_fsdp(opt) and opt['fp16_impl'] == 'safe'


def should_sync_gradnorm(opt):
    """
    Indicates whether fp16 optimizer wrappers should accumulate over workers.

    FP16 overflow detection and gradient clipping both require accumulating gradients
    across all workers when using FSDP, as workers only store a fraction of the
    gradients.
    """
    return (
        FSDP_AVAILABLE
        and opt['fp16']
        and opt.get('ddp_backend', DEFAULT_DDP_BACKEND) in ('zero2', 'zero3')
    )


def fsdp_wrap(module):
    """
    Helper function for wrapping the outermost root module.
    """
    return wrap(module)


def get_state_dict(model):
    """
    Get the state dict from the model.

    When using Pytorch FSDP, we can offload to CPU.
    """

    if FSDP_AVAILABLE:
        from torch.distributed.fsdp.fully_sharded_data_parallel import (
            FullStateDictConfig,
            StateDictType,
        )

        save_policy = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
        with FSDP.state_dict_type(model, StateDictType.FULL_STATE_DICT, save_policy):
            state = model.state_dict()
    else:
        state = model.state_dict()

    return state


@contextlib.contextmanager
def fsdp_join(*args):
    with Join([*args]):
        yield


class JoinableTrainLoop(TrainLoop, Joinable):
    """
    Joinable train loop.
    """

    def __init__(self, opt):
        import parlai.utils.distributed as dist_utils

        super().__init__(opt)
        self.__device = opt['gpu']
        self.__group = dist_utils.get_dist_group()

    def __call__(self):
        """
        Join caller.

        For now, don't do anything.
        """
        Join.notify_join_context(self)

    def join_hook(self, **kwargs) -> JoinHook:
        """
        Return our fake join hook.
        """
        return TrainLoopJoinHook(self)

    @property
    def join_device(self) -> torch.device:
        return self.__device

    @property
    def join_process_group(self):
        return self.__group


class TrainLoopJoinHook(JoinHook):
    """
    Join hook for train loop.

    Adapted from https://pytorch.org/tutorials/advanced/generic_join.html
    """

    def __init__(self, train_loop: JoinableTrainLoop):
        self.train_loop = train_loop

    def main_hook(self):
        pass

    def post_hook(self, is_last_joiner: bool):
        pass


class JoinableEvaluator(Evaluator, Joinable):
    """
    Joinable Evaluator.
    """

    def __init__(self, opt):
        import parlai.utils.distributed as dist_utils

        super().__init__(opt)
        self.__device = opt['gpu']
        self.__group = dist_utils.get_dist_group()

    def __call__(self):
        """
        Join caller.

        For now, don't do anything.
        """
        Join.notify_join_context(self)

    def join_hook(self, **kwargs) -> JoinHook:
        """
        Return our fake join hook.
        """
        return EvaluatorJoinHook(self)

    @property
    def join_device(self) -> torch.device:
        return self.__device

    @property
    def join_process_group(self):
        return self.__group


class EvaluatorJoinHook(JoinHook):
    def __init__(self, evaluator: JoinableEvaluator):
        self.evaluator = evaluator

    def main_hook(self):
        pass

    def post_hook(self, is_last_joiner: bool):
        pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Tests tod world + script, notably for batching, by comparing saved script logs to the
data that should have been generated.

Metrics are handled in separate files.
"""

import copy
import unittest

import parlai.core.tod.tod_test_utils.test_agents as test_agents
import parlai.core.tod.tod_core as tod_core
import parlai.scripts.tod_world_script as tod_world_script
from parlai.core.tod.tod_agents import StandaloneApiAgent


class TestTodWorldScript(tod_world_script.TodWorldScript):
    """
    Wrap around it to check its logic; also makes it easier to do things w/ underlying
    World.
    """

    def _get_tod_agents(self, opt):
        """
        Hack so we can separate out logic of making sure agent parsing is correct.
        """
        if hasattr(self, "agents"):
            return self.agents
        return super()._get_tod_agents(opt)

    def _save_outputs(self, opt, world, logger, episode_metrics):
        self.world = world
        self.logger = logger


class TodWorldInScriptTestBase(unittest.TestCase):
    def add_tod_world_opts(self, base_opts):
        """
        Convenience since we're initing the opt directly without parlai parser.
        """
        opts = copy.deepcopy(base_opts)
        opts["datatype"] = "DUMMY"
        opts["datafile"] = "DUMMY"
        opts["episodes_randomization_seed"] = -1
        opts["standalone_api_file"] = test_agents.API_DATABASE_FILE
        opts["exact_api_call"] = True
        opts["log_keep_fields"] = "all"
        opts["display_examples"] = False
        opts[
            "include_api_schemas"
        ] = True  # do this to test_agents.make sure they're done correctly.
        return opts

    def setup_agents(self, added_opts):
        full_opts = self.add_tod_world_opts(added_opts)
        sys = test_agents.ApiCallAndSysUttAgent(full_opts)
        agents = [
            test_agents.UserUttAgent(full_opts),
            sys,
            StandaloneApiAgent(full_opts),
            sys,
            test_agents.ApiSchemaAgent(full_opts),
            test_agents.GoalAgent(full_opts),
        ]
        return agents, full_opts

    def _test_roundDataCorrect(self):
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__SINGLE_API_CALL)
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__MULTI_ROUND)
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__MULTI_EPISODE)
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__MULTI_EPISODE_BS)

    def _check_correctness_from_script_logs(
        self, script, opt, process_round_utts=lambda x: x
    ):
        """
        Last argument is only relevant for the max_turn test.
        """
        max_rounds = opt[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        max_episodes = opt[test_agents.TEST_NUM_EPISODES_OPT_KEY]
        # there's something funky with logger.get_log() that inserts a space, but not gonna worry about it for now
        logs = [x for x in script.logger.get_logs() if len(x) > 0]
        for episode_idx in range(max_episodes):
            episode_from_world = logs[episode_idx]
            # first round is context
            context = episode_from_world[0]
            self.assertEquals(
                context[0]["text"],
                "APIS: "
                + tod_core.SerializationHelpers.list_of_maps_to_str(
                    test_agents.make_api_schemas_machine(max_rounds)
                ),
            )
            self.assertEquals(
                context[3]["text"],
                "GOAL: "
                + tod_core.SerializationHelpers.list_of_maps_to_str(
                    test_agents.make_goal_calls_machine(max_rounds)
                ),
            )
            # Check the rest
            world_utts = [[x["text"] for x in turn] for turn in episode_from_world[1:]]
            # ... ignore the last DONE turn here cause it's not that important

            self.assertEquals(
                world_utts[:-1],
                process_round_utts(
                    test_agents.get_round_utts(episode_idx, max_rounds)[:-1]
                ),
            )


class TodWorldSingleBatchTest(TodWorldInScriptTestBase):
    """
    Checks that saved data is correct with a single batch.
    """

    def _test_roundDataCorrect_helper(self, config):
        config["batchsize"] = 1
        config["max_turns"] = 10
        agents, opt = self.setup_agents(config)
        script = TestTodWorldScript(opt)
        script.agents = agents
        script.run()
        self._check_correctness_from_script_logs(script, opt)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()

    def test_max_turn(self):
        self._test_max_turn_helper(4)
        self._test_max_turn_helper(7)

    def _test_max_turn_helper(self, max_turns):
        config = {}
        config["batchsize"] = 1
        config["max_turns"] = max_turns
        config[test_agents.TEST_NUM_ROUNDS_OPT_KEY] = 10
        config[test_agents.TEST_NUM_EPISODES_OPT_KEY] = 5  # cause why not
        agents, opt = self.setup_agents(config)
        script = TestTodWorldScript(opt)
        script.agents = agents
        script.run()

        def filter_round_utt(utts):
            # tad imprecise, but more important that it does stop.
            # subtract 1 for the context turn, then 1 cause there's an off by one somewhere
            return utts[: max_turns - 2]

        self._check_correctness_from_script_logs(script, opt, filter_round_utt)


class TodWorldNonSingleBatchTest(TodWorldInScriptTestBase):
    """
    Checks saved data is correct with larger batchsizes.
    """

    def _test_roundDataCorrect_helper(self, config):
        config["batchsize"] = 4
        config["max_turns"] = 10
        agents, opt = self.setup_agents(config)
        script = TestTodWorldScript(opt)
        script.agents = agents
        script.run()
        self._check_correctness_from_script_logs(script, opt)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TodWorldTestSingleDumpAgents(TodWorldInScriptTestBase):
    """
    Just to be safe, make sure that the agents with "single" versions (ex goal + api
    schema) are correctly aligned.

    (Already tested in the agents test file as well, but to be safe.)
    """

    def setup_agents(self, added_opts, api_agent, goal_agent):
        full_opts = self.add_tod_world_opts(added_opts)
        full_opts["fixed_response"] = "USER: [DONE]"
        sys = test_agents.ApiCallAndSysUttAgent(full_opts)
        agents = [
            test_agents.UserUttAgent(full_opts),
            sys,
            StandaloneApiAgent(full_opts),
            sys,
            api_agent(full_opts),
            goal_agent(full_opts),
        ]
        return agents, full_opts

    def _test_SingleGoalApiResp_helper(self, batchsize, num_episodes):
        config = {}
        config["batchsize"] = batchsize
        config[test_agents.TEST_NUM_ROUNDS_OPT_KEY] = 10
        config[test_agents.TEST_NUM_EPISODES_OPT_KEY] = num_episodes
        single_agents, opt = self.setup_agents(
            config, test_agents.SingleApiSchemaAgent, test_agents.SingleGoalAgent
        )
        single_script = TestTodWorldScript(opt)
        single_script.agents = single_agents
        single_script.run()
        single_logs = [x for x in single_script.logger.get_logs() if len(x) > 0]

        multi_agents, opt = self.setup_agents(
            config, test_agents.ApiSchemaAgent, test_agents.GoalAgent
        )
        multi_script = TestTodWorldScript(opt)
        multi_script.agents = multi_agents
        multi_script.run()
        multi_logs = [x for x in single_script.logger.get_logs() if len(x) > 0]

        single_idx = 0
        for multi_log in multi_logs:
            context = multi_log[0]
            goals = tod_core.SerializationHelpers.str_to_goals(
                context[3]["text"][len("GOAL:") :].strip()
            )
            for goal in goals:
                single_context = single_logs[single_idx][0]
                single_goal = tod_core.SerializationHelpers.str_to_goals(
                    single_context[3]["text"][len("GOAL:") :].strip()
                )
                self.assertEqual(len(single_goal), 1)
                self.assertEquals(goal, single_goal[0])
                single_des = tod_core.SerializationHelpers.str_to_api_schemas(
                    single_context[0]["text"][len("APIS:") :].strip()
                )
                self.assertEqual(len(single_des), 1)
                self.assertEqual(single_goal[0]["api_name"], single_des[0]["api_name"])

                single_idx += 1

    def test_SingleGoalApiResp_helper_singleBatch(self):
        self._test_SingleGoalApiResp_helper(1, 2)
        self._test_SingleGoalApiResp_helper(1, 5)

    def test_SingleGoalApiResp_helper_multiBatch(self):
        self._test_SingleGoalApiResp_helper(4, 8)
        self._test_SingleGoalApiResp_helper(4, 11)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
from parlai.core.build_data import DownloadableFile

import os

RESOURCES = [
    DownloadableFile(
        url='1kVr-YsUVFisceiIklvpWEe0kHNSIFtNh',
        file_name='entailment_trees_emnlp2021_data_v3.zip',
        hashcode='fe05a02f181bb3d27fa2f8bafda824f7a988af9df59f848d694458925be7c497',
        zipped=True,
        from_google=True,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'entailment_bank')
    version = '1.0'

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print("[building data: " + dpath + "]")

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
Main launch script for single-host, multi-GPU training.

This is a drop-in replacement for [train_model]. This script will
launch N subprocess, each which runs the full training loop
independently.

Uses torch.nn.parallel.DistributedDataParallel for its main uses. Agents
must specifically implement the wrapper of DistributedDatParallel, but
all TorchRankerAgents and TorchGeneratorAgents support this.

## Examples

```shell
parlai multiprocessing_train -m transformer/generator --batchsize 16 --task convai2 --model-file /tmp/mymodel
```
"""

import torch
import os
import signal
import traceback
import parlai.scripts.train_model as single_train
import parlai.utils.distributed as distributed_utils
import parlai.utils.fsdp as fsdp_utils
from parlai.core.script import ParlaiScript, register_script


def multiprocess_train(
    rank, opt, port=61337, rank_offset=0, gpu=None, hostname='localhost'
):
    init_method = f"tcp://{hostname}:{port}"
    with distributed_utils.distributed_context(
        rank, opt, rank_offset, gpu, init_method=init_method
    ) as opt:
        # Run the actual training
        opt['multiprocessing'] = True
        loop = fsdp_utils.JoinableTrainLoop(opt)
        try:
            with fsdp_utils.fsdp_join(loop):
                return loop.train()
        except Exception:
            import parlai.utils.logging as logging

            logging.critical(traceback.format_exc())
            logging.critical(
                f"Got the above exception on worker {rank + rank_offset}. "
                "This may cause hangs requiring manual killing of processes."
            )
            raise


def launch_and_train(opt, port=None):
    """
    Perform a fork() to many processes.
    """
    if port is None:
        port = distributed_utils.find_free_port()
    # Launch multiple subprocesses
    spawncontext = torch.multiprocessing.start_processes(
        multiprocess_train,
        # need to give rank offset as 1 to cover the fact that the main
        # process is rank 0, but that spawn() doesn't let you control rank
        (opt, port, 1),
        nprocs=opt['distributed_world_size'] - 1,  # main proc will also run loop
        join=False,
        start_method='spawn',
    )

    try:
        retval = multiprocess_train(0, opt, port)
        spawncontext.join()
        return retval
    except KeyboardInterrupt:
        # tell the subprocesses to stop too
        for p in spawncontext.processes:
            if p.is_alive():
                os.kill(p.pid, signal.SIGINT)
        raise


def setup_args():
    parser = single_train.setup_args()
    parser.add_distributed_training_args()
    parser.set_defaults(distributed_world_size=torch.cuda.device_count())
    return parser


@register_script("multiprocessing_train", aliases=["mp_train"], hidden=True)
class MultiProcessTrain(ParlaiScript):
    @classmethod
    def setup_args(cls):
        argparser = setup_args()
        argparser.add_argument('--port', type=int, default=None)
        return argparser

    def run(self):
        if self.opt['port'] is None:
            port = None
        else:
            port = self.opt['port']
        return launch_and_train(self.opt, port)


if __name__ == '__main__':
    MultiProcessTrain.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from parlai.core.mutators import register_mutator, MessageMutator, ManyEpisodeMutator
from typing import Optional
from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
import parlai.tasks.wizard_of_internet.constants as CONST
from parlai.tasks.wizard_of_wikipedia.mutators import (
    AddLabel as AddLabelWizWiki,
    AddLabelLM as AddLabelLMWizWiki,
    CheckedSentenceAsLabel as CheckedSentenceAsLabelWizWiki,
    AddCheckedSentence as AddCheckedSentenceWizWiki,
)


@register_mutator("woi_add_checked_sentence_to_input")
class AddCheckedSentence(AddCheckedSentenceWizWiki):
    """
    Adds the checked sentence to the end of the text.

    E.g. run with: parlai display_data -t wizard_of_internet -n 100 -dt valid --mutators
    flatten,add_checked_sentence_to_input_woi
    """

    @property
    def checked_sentence_kword(self):
        return CONST.SELECTED_SENTENCES


@register_mutator("woi_checked_sentence_as_label")
class CheckedSentenceAsLabel(CheckedSentenceAsLabelWizWiki):
    """
    Uses the checked sentence (knowledge) as label.

    E.g. run with: parlai display_data -t wizard_of_internet -n 100 -dt valid --mutators
    flatten,checked_sentence_as_label_woi
    """

    @property
    def checked_sentence_kword(self):
        return CONST.SELECTED_SENTENCES


@register_mutator("woi_add_label_to_input")
class AddLabel(AddLabelWizWiki):
    """
    Adds the dialogue sentence to the input.

    E.g. run with: parlai display_data -t wizard_of_internet -n 100 -dt valid --mutators
    flatten,checked_sentence_as_label_woi,add_label_to_input_woi
    """

    pass


@register_mutator("woi_add_label_to_input_lm")
class AddLabelLM(AddLabelLMWizWiki):
    """
    Adds the dialogue sentence to the input (language modeling version).

    Language modeling version where a random piece of the label is sampled in
    the input. The rest is placed inside special tokens.

    E.g. run with: parlai display_data -t wizard_of_internet -n 100 -dt valid --mutators
    flatten,add_label_to_input_lm_woi

    To add the checked sentence as the label, use:
        parlai display_data -t wizard_of_internet -n 100 -dt valid --mutators
        flatten,add_label_to_input_lm_woi,checked_sentence_as_label_woi
    """

    pass


@register_mutator("woi_filter_no_passage_used")
class WoiFilterNoPassageUsed(ManyEpisodeMutator):
    """
    Allows to filter any examples where no passage was selected to base the wizard reply
    on.

    This works best in flattened mode. E.g. run with: parlai display_data -t
    wizard_of_internet -n 100 -dt valid --mutators flatten+filter_no_passage_used
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        for e in episode:
            checked_sentences = e.get(CONST.SELECTED_SENTENCES)
            checked_sentences = ' '.join(checked_sentences)
            if checked_sentences == CONST.NO_SELECTED_SENTENCES_TOKEN:
                pass
            else:
                out_episodes.append([e])
        return out_episodes


@register_mutator("woi_keep_only_no_passage_used")
class WoiKeepOnlyNoPassageUsed(ManyEpisodeMutator):
    """
    Filter all examples where passages are used.
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        for e in episode:
            checked_sentences = e.get(CONST.SELECTED_SENTENCES)
            checked_sentences = ' '.join(checked_sentences)
            if checked_sentences == CONST.NO_SELECTED_SENTENCES_TOKEN:
                out_episodes.append([e])
        return out_episodes


@register_mutator("woi_filter_selected_knowledge_in_retrieved_docs")
class WoiFilterSelectedKnowledgeInRetrievedDocs(ManyEpisodeMutator):
    """
    Allows to filter any examples where '__retrieved-docs__' field doesn't contain the
    '__selected-sentences__'.
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        for e in episode:
            checked_sentences = e.get(
                CONST.SELECTED_SENTENCES,
                e.get('labels', [CONST.NO_SELECTED_SENTENCES_TOKEN]),
            )
            docs = ' '.join(e.get('__retrieved-docs__'))
            if ' '.join(checked_sentences) != CONST.NO_SELECTED_SENTENCES_TOKEN:
                found = True
                for sent in checked_sentences:
                    s = sent.lstrip(' ').rstrip(' ')
                    if s not in docs:
                        found = False
                if found:
                    out_episodes.append([e])
            else:
                pass
        return out_episodes


def chunk_docs_in_message(message, chunk_sz):
    if CONST.RETRIEVED_DOCS not in message:
        return message
    new_message = message.copy()
    docs = message[CONST.RETRIEVED_DOCS]
    titles = message.get(CONST.RETRIEVED_DOCS_TITLES)
    urls = message.get(CONST.RETRIEVED_DOCS_URLS)
    new_docs = []
    new_titles = []
    new_urls = []
    checked_sentences = list(
        message.get(
            CONST.SELECTED_SENTENCES,
            message.get('labels', [CONST.NO_SELECTED_SENTENCES_TOKEN]),
        )
    )
    for i in range(len(checked_sentences)):
        checked_sentences[i] = checked_sentences[i].lstrip(' ').rstrip(' ')
    if ' '.join(checked_sentences) == CONST.NO_SELECTED_SENTENCES_TOKEN:
        checked_sentences = []
    for ind in range(len(docs)):
        d = docs[ind]
        # Guarantees that checked sentences are not split in half (as we split by space).
        for i in range(len(checked_sentences)):
            d = d.replace(checked_sentences[i], "||CHECKED_SENTENCE_{i}||")
        while True:
            end_chunk = d.find(' ', chunk_sz)
            if end_chunk == -1:
                # last chunk
                for i in range(len(checked_sentences)):
                    d = d.replace("||CHECKED_SENTENCE_{i}||", checked_sentences[i])
                new_docs.append(d)
                new_titles.append(titles[ind])
                new_urls.append(urls[ind])
                break
            else:
                new_d = d[0:end_chunk]
                for i in range(len(checked_sentences)):
                    new_d = new_d.replace(
                        "||CHECKED_SENTENCE_{i}||", checked_sentences[i]
                    )
                new_docs.append(new_d)
                new_titles.append(titles[ind])
                new_urls.append(urls[ind])
                d = d[end_chunk + 1 : -1]
    new_message.force_set(CONST.RETRIEVED_DOCS, new_docs)
    new_message.force_set(CONST.RETRIEVED_DOCS_TITLES, new_titles)
    new_message.force_set(CONST.RETRIEVED_DOCS_URLS, new_urls)
    return new_message


@register_mutator("woi_chunk_retrieved_docs")
class WoiChunkRetrievedDocs(MessageMutator):
    """
    Chunks '__retrieved-docs__' into smaller docs (max 100 words each).
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser.add_argument(
            '--woi-doc-chunk-size',
            default=500,
            type=int,
            help='Document chunk size (in characters).',
        )

    def message_mutation(self, message: Message) -> Message:
        chunk_sz = self.opt.get('woi_doc_chunk_size', 500)
        return chunk_docs_in_message(message, chunk_sz)


@register_mutator("woi_dropout_retrieved_docs")
class WoiDropoutRetrievedDocs(MessageMutator):
    """
    Drops out '__retrieved-docs__' to only keep a maximum number in each example.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser.add_argument(
            '--woi-doc-max-chunks',
            default=100,
            type=int,
            help='Largest number of chunks to use, others will be dropped out at random. Chunks containing gold checked sentences will not be removed.',
        )

    def message_mutation(self, message: Message) -> Message:
        if CONST.RETRIEVED_DOCS not in message:
            return message
        new_message = message.copy()
        docs = message.get(CONST.RETRIEVED_DOCS)
        new_docs = []
        max_chunks = self.opt.get('woi_doc_max_chunks', 100)
        if max_chunks > 0:
            keep = torch.randperm(len(docs))[0:max_chunks]
        else:
            keep = torch.randperm(len(docs))
        remove = torch.ones(len(docs))
        remove[keep] = 0

        for i in range(len(docs)):
            if remove[i] == 0:
                new_docs.append(docs[i])
            else:
                # We may still keep the doc if it contains the gold checked sentence(s).
                checked_sentences = message.get(
                    CONST.SELECTED_SENTENCES,
                    message.get('labels', [CONST.NO_SELECTED_SENTENCES_TOKEN]),
                )
                d = docs[i]
                found = False
                if ' '.join(checked_sentences) != CONST.NO_SELECTED_SENTENCES_TOKEN:
                    for sent in checked_sentences:
                        s = sent.lstrip(' ').rstrip(' ')
                        if s in d:
                            found = True
                if found:
                    new_docs.append(docs[i])

        new_message.force_set(CONST.RETRIEVED_DOCS, new_docs)
        return new_message



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Classifier Re-Ranker Object.

Provided with a classifier model file, the re-ranker provides an API for re-ranking
candidate outputs based on maximizing the probability of a given provided class.
"""
from typing import Optional, List
from parlai.core.agents import create_agent_from_model_file
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser

from parlai.agents.reranker.reranker import (
    AbstractReranker,
    AbstractGeneratorRerankAgent,
)


class ClassifierReranker(AbstractReranker):
    def __init__(self, opt: Opt, shared=None):
        """
        Initializes reranker.
        """
        super().__init__(opt, shared)
        self.include_label_cand_only = opt['include_label_cand_only']
        self.target_label = opt['target_label']

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None):
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        classifier_reranker = parser.add_argument_group('ClassifierReranker args')
        classifier_reranker.add_argument(
            '--include-label-cand-only',
            type='bool',
            default=False,
            help='When passing inputs to the classifier, use only the label targets if set to True.',
        )
        classifier_reranker.add_argument(
            '--target-label',
            type=str,
            default='pos',
            help='The name of the target class/label that we want to maximize the probability of.',
        )
        return parser

    def init_predictor(self, opt: Opt, shared=None):
        if not shared:
            override = {
                'return_cand_scores': True,
                'datatype': 'valid',
                'interactive_mode': opt.get('interactive_mode', True),
                'ignore_bad_candidates': True,
                'encode_candidate_vecs': True,
                'interactive_candidates': 'inline',
            }  # to not init optim
            self.predictor = create_agent_from_model_file(
                self.predictor_model_file, opt_overrides=override
            )
        else:
            self.predictor = shared['predictor']

    def batch_classify(
        self, contexts: List[str], predictor_label_candidates: List[str]
    ) -> Message:
        """
        Use predictor to predict given augmented context.

        :param context:
            augmented context with response candidates
        :param predictor_label_candidates:
            optional array of label candidates to pass to the predictor
        :return output:
            return output from ranker act
        """
        cands = []
        if self.include_label_cand_only:
            for c in predictor_label_candidates:
                cands.append(c)
        else:
            cands = [
                context + self.delimiter + response
                for context, response in zip(contexts, predictor_label_candidates)
            ]
        predictor_outputs = self.batch_predict(cands, self.predictor.class_list)
        return predictor_outputs

    def get_class_to_rerank_for(
        self, observation: Message, full_context: str
    ) -> Optional[str]:
        return self.target_label

    def is_context(self, utt: str) -> bool:
        return False

    def get_predictor_label_candidates(
        self, observation: Message, context: str
    ) -> List[str]:
        return self.predictor.class_list


class ClassifierRerankerAgent(AbstractGeneratorRerankAgent):
    """
    Generative Re-rank agent for adding a ClassifierReranker.
    """

    @classmethod
    def get_reranker_class(cls):
        return ClassifierReranker



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import hashlib
import json
import os
import random

import numpy as np
import parlai.core.build_data as build_data
import parlai.utils.logging as logging
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/dialogue_safety/single_turn_safety.json',
        'single_turn_safety.json',
        'f3a46265aa639cfa4b55d2be4dca4be1c596acb5e8f94d7e0041e1a54cedd4cd',
        zipped=False,
    ),
]
# The size of safety mix
DATA_SIZE = {'train': 200, 'valid': 24, 'test': 24}

TROLL_TYPES = {
    'troll',
    'master_troll',
    'lazy_troll',
    'safe_troll',
    'unsafe_troll',
    'gaslight_troll',
}

FILE_TYPE_EXTENSIONS = {
    'train': '_train.jsonl',
    'valid': '_valid.jsonl',
    'test': '_test.jsonl',
}


# Dataset analysis
# check how many samples from standard vs adversarial
def analysis_dataset(dataset):
    standard_sample_cnt = 0
    adversarial_sample_cnt = 0
    flipped_label_cnt = 0
    duplicated_samples = {}
    for each_sample in dataset:
        if each_sample['source'] == 'standard':
            standard_sample_cnt += 1
        else:
            adversarial_sample_cnt += 1
        if each_sample['gold_labels'][0] != each_sample['labels'][0]:
            flipped_label_cnt += 1
        if each_sample['text'] in duplicated_samples:
            duplicated_samples[each_sample['text']] += 1
        else:
            duplicated_samples[each_sample['text']] = 1
    duplicated_samples = {
        k: v for k, v in sorted(duplicated_samples.items(), key=lambda item: item[1])
    }
    duplicated_samples = {k: v for k, v in duplicated_samples.items() if int(v) >= 2}

    all_values = list(duplicated_samples.values())
    max_value = 0 if len(all_values) == 0 else max(all_values)
    mean_value = 0 if len(all_values) == 0 else np.mean(all_values)
    logging.info(
        f'Real noise level {flipped_label_cnt / (standard_sample_cnt + adversarial_sample_cnt)}'
    )
    logging.info(
        f'Standard samples: {standard_sample_cnt} adversarial samples: {adversarial_sample_cnt}'
    )
    logging.info(
        f'It has {len(duplicated_samples)} duplicated samples, the the maximum appearance {max_value}, in average {mean_value}'
    )


def flip_labels(samples, N):
    flip_cnt = 0
    for sample in samples:
        text = sample['text']
        flip = hashlib.md5(text.encode('utf8')).digest()[0] < N * 256
        labels = sample['labels']
        # deterministic flipping for each example
        if flip:
            flip_cnt += 1
            if labels[0] == '__ok__':
                new_labels = ['__notok__']
            elif labels[0] == '__notok__':
                new_labels = ['__ok__']
            else:
                raise ValueError('labels are not in the right format.')
        if flip:
            sample['labels'] = new_labels
    logging.debug(f'{flip_cnt} labels flipped.')
    return samples


def sample_from_single_dataset(
    dataset_for_sample, source, utterance_num, user_id_to_start, random_int_for_sample
):
    utterance_cnt = 0
    user_id = user_id_to_start
    output_data = []
    while utterance_cnt < utterance_num:
        #     draw a number for user id
        current_user_utterance_num = int(random.choice(random_int_for_sample))
        if current_user_utterance_num + utterance_cnt > utterance_num:
            current_user_utterance_num = utterance_num - utterance_cnt
        utterance_cnt += current_user_utterance_num
        #     sample with replacement
        samples = copy.deepcopy(
            random.choices(dataset_for_sample, k=current_user_utterance_num)
        )
        for each_sample in samples:
            #         adding user id
            each_sample['user_id'] = user_id
            each_sample['source'] = source
            each_sample['gold_labels'] = copy.deepcopy(each_sample['labels'])
            output_data.append(each_sample)
        user_id += 1
    assert len(output_data) == utterance_num
    return output_data, user_id


def sample_from_both_datasets(
    standard_dataset_for_sample,
    adversarial_dataset_for_sample,
    utterance_num,
    user_id_to_start,
    random_int_for_sample,
):
    utterance_cnt = 0
    user_id = user_id_to_start
    output_data = []
    while utterance_cnt < utterance_num:
        #     draw a number for user id
        current_user_utterance_num = int(random.choice(random_int_for_sample))
        if current_user_utterance_num + utterance_cnt > utterance_num:
            current_user_utterance_num = utterance_num - utterance_cnt
        utterance_cnt += current_user_utterance_num
        #     sample with replacement
        #     flip a coin
        samples = []
        for _ in range(current_user_utterance_num):
            if random.random() > 0.5:
                sample = copy.deepcopy(random.choice(standard_dataset_for_sample))
                sample['source'] = 'standard'
            else:
                sample = copy.deepcopy(random.choice(adversarial_dataset_for_sample))
                sample['source'] = 'adversarial'
            samples.append(sample)
        for each_sample in samples:
            #         adding user id
            each_sample['user_id'] = user_id
            each_sample['gold_labels'] = copy.deepcopy(each_sample['labels'])
            output_data.append(each_sample)
        user_id += 1
    assert len(output_data) == utterance_num
    return output_data, user_id


def set_labels(samples, label):
    for sample in samples:
        sample['labels'] = [label]
    return samples


def set_user_type(samples, user_type):
    for sample in samples:
        sample['user_type'] = user_type
    return samples


def set_random_labels(samples, N):
    random_cnt = 0
    for sample in samples:
        text = sample['text']
        change_label = hashlib.md5(text.encode('utf8')).digest()[0] < N * 256
        # deterministic flipping for each example
        if change_label:
            random_cnt += 1
        if random.random() > 0.5:
            sample['labels'] = ['__ok__']
        else:
            sample['labels'] = ['__notok__']
    logging.debug(f'{random_cnt} labels randomized.')
    return samples


def generate_user_based_dataset(
    N, generation_fold, safety_dataset, user_ratios, random_int_for_sample
):
    total_number_of_samples = DATA_SIZE[generation_fold]
    # Take the first 800 samples from safety set to make it balanced between good and bad.
    standard_safe = safety_dataset['standard'][generation_fold]['1']['good'][:800]
    standard_unsafe = safety_dataset['standard'][generation_fold]['1']['bad'][:800]
    adversarial_safe = safety_dataset['adversarial'][generation_fold]['1']['good'][:800]
    adversarial_unsafe = safety_dataset['adversarial'][generation_fold]['1']['bad'][
        :800
    ]
    standard_dataset_for_sample = random.sample(
        standard_safe + standard_unsafe, len(standard_safe + standard_unsafe)
    )
    adversarial_dataset_for_sample = random.sample(
        adversarial_safe + adversarial_unsafe,
        len(adversarial_safe + adversarial_unsafe),
    )
    user_id = 0

    # for helpers
    helper_utterance_num = int(total_number_of_samples * user_ratios['helper'])
    helper_samples, user_id = sample_from_single_dataset(
        standard_dataset_for_sample,
        'standard',
        helper_utterance_num,
        user_id,
        random_int_for_sample,
    )
    helper_samples = set_user_type(helper_samples, 'helper')

    # for master helper
    master_helper_utterance_num = int(
        total_number_of_samples * user_ratios['master_helper']
    )
    master_helper_samples, user_id = sample_from_single_dataset(
        adversarial_dataset_for_sample,
        'adversarial',
        master_helper_utterance_num,
        user_id,
        random_int_for_sample,
    )
    master_helper_samples = set_user_type(master_helper_samples, 'master_helper')

    # for trolls
    troll_utterance_num = int(total_number_of_samples * user_ratios['troll'])
    troll_samples, user_id = sample_from_single_dataset(
        standard_dataset_for_sample,
        'standard',
        troll_utterance_num,
        user_id,
        random_int_for_sample,
    )
    # flip samples for trolls
    troll_samples = flip_labels(troll_samples, N)
    troll_samples = set_user_type(troll_samples, 'troll')

    # sample master trolls
    master_troll_utterance_num = int(
        total_number_of_samples * user_ratios['master_troll']
    )
    master_troll_samples, user_id = sample_from_single_dataset(
        adversarial_dataset_for_sample,
        'adversarial',
        master_troll_utterance_num,
        user_id,
        random_int_for_sample,
    )
    # flip samples for master trolls
    master_troll_samples = flip_labels(master_troll_samples, N)
    master_troll_samples = set_user_type(master_troll_samples, 'master_troll')

    # sample safe trolls
    safe_troll_utterance_num = int(total_number_of_samples * user_ratios['safe_troll'])
    safe_troll_samples, user_id = sample_from_both_datasets(
        standard_dataset_for_sample,
        adversarial_dataset_for_sample,
        safe_troll_utterance_num,
        user_id,
        random_int_for_sample,
    )
    safe_troll_samples = set_labels(safe_troll_samples, '__ok__')
    safe_troll_samples = set_user_type(safe_troll_samples, 'safe_troll')

    # sample unsafe trolls
    unsafe_troll_utterance_num = int(
        total_number_of_samples * user_ratios['unsafe_troll']
    )
    unsafe_troll_samples, user_id = sample_from_both_datasets(
        standard_dataset_for_sample,
        adversarial_dataset_for_sample,
        unsafe_troll_utterance_num,
        user_id,
        random_int_for_sample,
    )
    unsafe_troll_samples = set_labels(unsafe_troll_samples, '__notok__')
    unsafe_troll_samples = set_user_type(unsafe_troll_samples, 'unsafe_troll')

    # happy helper
    happy_helper_utterance_num = int(
        total_number_of_samples * user_ratios['happy_helper']
    )
    happy_helper_samples, user_id = sample_from_single_dataset(
        random.sample(standard_safe, len(standard_safe)),
        'standard',
        happy_helper_utterance_num,
        user_id,
        random_int_for_sample,
    )
    happy_helper_samples = set_user_type(happy_helper_samples, 'happy_helper')

    # gaslight trolls
    gaslight_troll_utterance_num = int(
        total_number_of_samples * user_ratios['gaslight_troll']
    )
    gaslight_troll_samples, user_id = sample_from_single_dataset(
        random.sample(adversarial_unsafe, len(adversarial_unsafe)),
        'adversarial',
        gaslight_troll_utterance_num,
        user_id,
        random_int_for_sample,
    )
    gaslight_troll_samples = set_labels(gaslight_troll_samples, '__ok__')
    gaslight_troll_samples = set_user_type(gaslight_troll_samples, 'gaslight_troll')

    # lazy trolls
    lazy_troll_utterance_num = int(total_number_of_samples * user_ratios['lazy_troll'])
    lazy_troll_samples, user_id = sample_from_both_datasets(
        standard_dataset_for_sample,
        adversarial_dataset_for_sample,
        lazy_troll_utterance_num,
        user_id,
        random_int_for_sample,
    )
    lazy_troll_samples = set_random_labels(lazy_troll_samples, N)
    lazy_troll_samples = set_user_type(lazy_troll_samples, 'lazy_troll')

    dataset = (
        helper_samples
        + master_helper_samples
        + troll_samples
        + master_troll_samples
        + safe_troll_samples
        + unsafe_troll_samples
        + happy_helper_samples
        + gaslight_troll_samples
        + lazy_troll_samples
    )
    return dataset


def save_dataset_for_two_class(dataset, troll_type, output_data_path, generation_fold):
    # split dataset into postive and negative for balanced training
    output_file_name = os.path.join(
        output_data_path, troll_type + FILE_TYPE_EXTENSIONS[generation_fold]
    )
    with open(output_file_name, 'w') as outfile:
        for json_blob in dataset:
            outfile.write(json.dumps(json_blob))
            outfile.write("\n")

    dataset_pos = []
    dataset_neg = []
    for each_sample in dataset:
        if each_sample['labels'][0] == '__ok__':
            dataset_pos.append(each_sample)
        else:
            dataset_neg.append(each_sample)
    output_file_name = os.path.join(
        output_data_path, 'pos_' + troll_type + FILE_TYPE_EXTENSIONS[generation_fold]
    )
    with open(output_file_name, 'w') as outfile:
        for json_blob in dataset_pos:
            outfile.write(json.dumps(json_blob))
            outfile.write("\n")
    output_file_name = os.path.join(
        output_data_path, 'neg_' + troll_type + FILE_TYPE_EXTENSIONS[generation_fold]
    )
    with open(output_file_name, 'w') as outfile:
        for json_blob in dataset_neg:
            outfile.write(json.dumps(json_blob))
            outfile.write("\n")


def build(datapath):
    version = 'v1.0'
    dpath = os.path.join(datapath, 'safety_mix')

    if not build_data.built(dpath, version):
        logging.info('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)
        random.seed(42)
        np.random.seed(42)
        # The number of user utterances for sampling
        random_nums = np.random.normal(loc=10, scale=2, size=100000)
        random_int_for_sample = np.round(random_nums)
        # load safety dataset
        with open(os.path.join(dpath, 'single_turn_safety.json')) as safety_file:
            safety_data = json.load(safety_file)
        for troll_type in TROLL_TYPES:
            # by default we do a 50% 50% mix with helpers and trolls
            ratios = {
                'helper': 0.5,
                'troll': 0.0,
                'master_helper': 0.0,
                'master_troll': 0.0,
                'lazy_troll': 0.0,
                'safe_troll': 0.0,
                'unsafe_troll': 0.0,
                'gaslight_troll': 0.0,
                'happy_helper': 0.0,
            }
            # Set the troll type 50% accordingly
            ratios[troll_type] = 0.5
            # Set the noise ratio N for some type of trolls
            N = 0.8
            for generation_fold in ['train', 'valid', 'test']:
                # for deterministic data generation
                random.seed(42)
                np.random.seed(42)
                dataset = generate_user_based_dataset(
                    N, generation_fold, safety_data, ratios, random_int_for_sample
                )
                if generation_fold == 'train':
                    logging.info(f'Stats for troll type {troll_type}')
                    analysis_dataset(dataset)
                #         save_dataset(dataset)
                save_dataset_for_two_class(dataset, troll_type, dpath, generation_fold)
                # Mark the data as built.
                build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Blender 2.7B model fine-tuned on the SaFeRDialogues and BST (without persona) tasks.
"""

from parlai.core.build_data import download_models


def download(datapath):
    opt = {'datapath': datapath}
    version = 'v0.2'
    fnames = [f'models_{version}.tar.gz']
    download_models(
        opt,
        fnames,
        model_folder='saferdialogues',
        version=version,
        use_model_type=False,
    )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest  # noqa: F401


class TestDefaultTeacher(AutoTeacherTest):
    task = 'friends'


class TestAllCharactersTeacher(AutoTeacherTest):
    task = 'friends:all_characters'


class TestRachelTeacher(AutoTeacherTest):
    task = 'friends:rachel'


class TestMonicaTeacher(AutoTeacherTest):
    task = 'friends:monica'


class TestPhoebeTeacher(AutoTeacherTest):
    task = 'friends:phoebe'


class TestJoeyTeacher(AutoTeacherTest):
    task = 'friends:joey'


class TestChandlerTeacher(AutoTeacherTest):
    task = 'friends:chandler'


class TestRossTeacher(AutoTeacherTest):
    task = 'friends:ross'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

ROOT_URL = (
    "https://github.com/google-research-datasets/Taskmaster/raw/master/TM-3-2020/"
)

RESOURCES = [
    # raw data files
    DownloadableFile(
        f"{ROOT_URL}/ontology/apis.json",
        "apis.json",
        "7fe0545dece5dce292c19cc21b02afb0b88fb02a7929d2f8605fe4d6d965a6e0",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_00.json",
        "data_00.json",
        "4edf97557e1aa7f654bf97994b5eae42653ef6d4f5e50136f7664d7e5cdff7b6",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_01.json",
        "data_01.json",
        "a4fef75ec7824bb3fb29b8afe9ed63f3354b81077bb9b6f58f2349bb60d660fc",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_02.json",
        "data_02.json",
        "25cb4788c4c857740152612397b0e687dfeade14c7dbfeed501a53290725aedf",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_03.json",
        "data_03.json",
        "5e1f15544f259c804438458f34a6c7f00400e30dba7648c37a0b31886c5dd903",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_04.json",
        "data_04.json",
        "17d8deb3bc6c3551c564cb9003a47202a835e011477b1e3c81711b083bf4064b",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_05.json",
        "data_05.json",
        "12e58018be7b42aa16968ab914d03c56af9fae9fb5feec7448e01f738c50d2fe",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_06.json",
        "data_06.json",
        "a4e290dafc6362ca8d4d786d6e70f6339297b0414cca4fe0ec79c1d383d65673",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_07.json",
        "data_07.json",
        "becec1c0e9b88781fbb81f5d032826f85d77756cfae7c37acbca9b4b2a470fdb",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_08.json",
        "data_08.json",
        "c2718197cedfb1ff7b830492ddbdc36c213ac98c83ba974ff0ea0eeeffdfe9b5",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_09.json",
        "data_09.json",
        "a8867eb49010fa735b3c0b2f47df4cc225c79b6ca919009db857f454cccc986d",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_10.json",
        "data_10.json",
        "88a8ce81f2c843c7a52a1a6c016d790e6547d5a3ffdd000ca00f0676c02e077b",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_11.json",
        "data_11.json",
        "10e5d2dadb7a55b08d41605498b46320fa2df7b87c71a820df4245e9eeeba4d1",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_12.json",
        "data_12.json",
        "be255d29439d10fc33bb222a1294da0af7ed4bdf8ede84c389d9b744e9cdf343",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_13.json",
        "data_13.json",
        "a477d4668398b86bc7f68f86c20a423e81b4dbc41002a5d605dee0c6cee250b8",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_14.json",
        "data_14.json",
        "92c7a0ca1fce8845610a7fdb00972750228351b2c2065e76120c328c806cc1e5",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_15.json",
        "data_15.json",
        "e67e9d649d29f3edc12796cb4b1308cc7cb89f676c12e96d8fea558510b6ee18",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_16.json",
        "data_16.json",
        "f0d3f22ee8b9224ea4efc57182f1620f849f3b90050cc04992319aab04e7b453",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_17.json",
        "data_17.json",
        "15c469a030aae16a98969675b238f368436ec0a5074f3eb79458fdc81cf489d4",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_18.json",
        "data_18.json",
        "28dfe3f65c1cd7bab8ed9da0e667ee1aec1561cf41237531f30c889cafb0997d",
        zipped=False,
    ),
    DownloadableFile(
        f"{ROOT_URL}/data/data_19.json",
        "data_19.json",
        "7ae99d1237c25519269c3e6969ab0ecd88509828c436cbecb8a86831007c3868",
        zipped=False,
    ),
]


def build(opt):
    # get path to data directory
    dpath = os.path.join(opt["datapath"], "taskmaster-3")
    # define version if any
    version = "1.0"

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print("[building data: " + dpath + "]")

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.opt import Opt
import parlai.utils.logging as logging

from parlai.tasks.msc.agents import PersonaSummaryTeacher
from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin


class MscMemoryGenerationTeacher(BB3TeacherMixin, PersonaSummaryTeacher):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['include_last_session'] = True
        opt['nopersona_subsampling_weight'] = 0.05
        mutators = '+'.join(
            [
                'skip_retrieval_mutator',
                'prompt_memory_mutator',
                'prefix_speakers_opt',
                'format_gen_tasks_for_decoder_only',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Imported from https://allenai.org/data/entailmentbank
Paper: "Explaining Answers with Entailment Trees" (Dalvi et al 2021) https://aclanthology.org/2021.emnlp-main.585.pdf

2k multi-step entailment trees, explaining the answers to ARC science questions.
"""

import json
import os
import re

from parlai.utils.data import DatatypeHelper

from .build import build
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.utils.io import PathManager
from typing import List, Optional

from tasks.reasoning.agents import StepByStepReasoningTeacher
import random


class EntailmentBankStepByStepReasoningTeacher(StepByStepReasoningTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Solve the following entailment problem. Show your work."
        )
        parser.add_argument(
            "--task-id",
            type=str,
            default="1",
            choices=["1", "2", "3"],
            help="Use data from Task 1, 2, or 3 as defined in the paper",
        )
        parser.add_argument(
            "--extrinsic-step",
            type=bool,
            default=False,
        )
        return parser

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)  # NOTE: the call to build here
        opt['datafile'] = os.path.join(opt['datapath'], 'entailment_bank')
        self.id = 'entailment_bank'
        self.fold = DatatypeHelper.fold(opt['datatype'])
        self.entailment_random = random.Random(42)
        super().__init__(opt, shared)

    def get_data_for_fold(self, fold):
        extrinsic_step = self.opt.get("extrinsic_step", False)
        task_id = self.opt.get("task_id", "1")
        suffix = self.fold
        if self.fold == "valid":
            suffix = "dev"
        filename = os.path.join(
            self.opt['datafile'],
            f"entailment_trees_emnlp2021_data_v3/dataset/task_{task_id}/{suffix}.jsonl",
        )

        data = []
        with PathManager.open(filename) as f:
            for x in f.readlines():
                example = json.loads(x)
                context = example['context']
                context = self._parse_context(context)
                # question with context
                question = context + ". " + example['question']
                proof = example['full_text_proof']
                steps = self._get_steps(proof)
                answer = example['answer']
                m = {
                    "question": question,
                    "answer": answer,
                    "steps": steps,
                }
                if not extrinsic_step:
                    yield m
                else:
                    data.append(m)

        if extrinsic_step:
            for line in data:
                rand_steps = self.entailment_random.choice(data)["steps"]
                random_step = self.entailment_random.choice(rand_steps)
                yield {
                    "question": line["question"],
                    "answer": line["answer"],
                    "steps": line["steps"],
                    "extrinsic_step": random_step,
                }

    def _parse_context(self, cont: str) -> str:
        """
        Change format.

        Input: sent1: xxx sent2: yyy sent3: zzz
        Output: xxx. yyy. zzz.
        """
        parsed = re.split(r" ?sent\d+: ", cont)[1:]
        return ". ".join(parsed)

    def _get_steps(self, proof: str) -> List[str]:
        """
        Input:
            proof: " [BECAUSE] earth is a kind of celestial object [AND] a star is a kind of celestial
        object / celestial body [AND] apparent motion is when an object appears to move relative to
        another object 's position [INFER] int1: apparent motion of stars is when stars appear to move
        relative to earth's position [BECAUSE] int1 [AND] the earth rotating on its axis causes stars
        to appear to move across the sky at night [INFER] int2: the earth rotating on its axis causes
        apparent motion of stars [BECAUSE] int2 [AND] stars appear to move relative to the horizon
        during the night [INFER] int3: the earth rotating on its axis causes stars to move relative to
        the horizon during the night"
        Return:
            ['earth is a kind of celestial object', 'a star is a kind of celestial object / celestial
            body', "apparent motion is when an object appears to move relative to another object 's
            position", "Therefore apparent motion of stars is when stars appear to move relative to
            earth's position", 'the earth rotating on its axis causes stars to appear to move across
            the sky at night', 'Therefore the earth rotating on its axis causes apparent motion of
            stars', 'stars appear to move relative to the horizon during the night', 'Therefore the
            earth rotating on its axis causes stars to move relative to the horizon during the night']
        """
        steps = re.split(
            r" \[BECAUSE\] int\d+ \[AND\] | \[BECAUSE\] | \[AND\] ", proof
        )[1:]
        out = []
        for step in steps:
            sub_step = re.sub(r"int\d+\:", "Therefore", step)
            out.extend(sub_step.split(" [INFER] "))
        return out


class DefaultTeacher(EntailmentBankStepByStepReasoningTeacher):
    pass
